# -*- coding: utf-8 -*-
"""Tests for :mod:`orion.algo.asha`."""
import hashlib
import itertools
import logging

import numpy as np
import pytest

from orion.algo.asha import ASHA, ASHABracket, compute_budgets
from orion.algo.space import Fidelity, Integer, Real, Space
from orion.testing.algo import BaseAlgoTests


@pytest.fixture
def space():
    """Create a Space with a real dimension and a fidelity value."""
    space = Space()
    space.register(Real("lr", "uniform", 0, 1))
    space.register(Fidelity("epoch", 1, 9, 3))
    return space


@pytest.fixture
def b_config(space):
    """Return a configuration for a bracket."""
    fidelity_dim = space.values()[0]
    num_rungs = 3
    budgets = np.logspace(
        np.log(fidelity_dim.low) / np.log(fidelity_dim.base),
        np.log(fidelity_dim.high) / np.log(fidelity_dim.base),
        num_rungs,
        base=fidelity_dim.base,
    )
    return {
        "reduction_factor": fidelity_dim.base,
        "budgets": list(zip(list(map(int, budgets[::-1])), budgets)),
    }


@pytest.fixture
def asha(b_config, space):
    """Return an instance of ASHA."""
    return ASHA(space)


@pytest.fixture
def bracket(b_config, asha):
    """Return a `ASHABracket` instance configured with `b_config`."""
    return ASHABracket(asha, b_config["budgets"], 1)


@pytest.fixture
def rung_0():
    """Create fake points and objectives for rung 0."""
    points = np.linspace(0, 1, 9)
    return dict(
        n_trials=9,
        resources=1,
        results={
            hashlib.md5(str([point]).encode("utf-8")).hexdigest(): (point, (1, point))
            for point in points
        },
    )


@pytest.fixture
def rung_1(rung_0):
    """Create fake points and objectives for rung 1."""
    return dict(
        n_trials=9,
        resources=3,
        results={
            hashlib.md5(str([value[0]]).encode("utf-8")).hexdigest(): value
            for value in map(
                lambda v: (v[0], (3, v[0])), sorted(rung_0["results"].values())
            )
        },
    )


@pytest.fixture
def rung_2(rung_1):
    """Create fake points and objectives for rung 1."""
    return dict(
        n_trials=9,
        resources=9,
        results={
            hashlib.md5(str([value[0]]).encode("utf-8")).hexdigest(): value
            for value in map(
                lambda v: (v[0], (9, v[0])), sorted(rung_1["results"].values())
            )
        },
    )


def force_observe(asha, point, results):

    full_id = asha.get_id(point, ignore_fidelity=False)
    asha.register(point, results)

    bracket = asha._get_bracket(point)
    id_wo_fidelity = asha.get_id(point, ignore_fidelity=True)
    asha.trial_to_brackets[id_wo_fidelity] = bracket

    asha.observe([point], [results])


def test_compute_budgets():
    """Verify proper computation of budgets on a logarithmic scale"""
    # Check typical values
    assert compute_budgets(1, 16, 4, 3, 1) == [[(16, 1), (4, 4), (1, 16)]]
    # Check rounding (max_resources is not a multiple of reduction_factor)
    assert compute_budgets(1, 30, 4, 3, 1) == [[(16, 1), (4, 5), (1, 30)]]
    # Check rounding (min_resources may be rounded below its actual value)
    assert compute_budgets(25, 1000, 2, 6, 1) == [
        [(32, 25), (16, 52), (8, 109), (4, 229), (2, 478), (1, 1000)]
    ]
    # Check min_resources
    assert compute_budgets(5, 125, 5, 3, 1) == [[(25, 5), (5, 25), (1, 125)]]
    # Check num_rungs
    assert compute_budgets(1, 16, 2, 5, 1) == [
        [(16, 1), (8, 2), (4, 4), (2, 8), (1, 16)]
    ]


def test_compute_compressed_budgets():
    """Verify proper computation of budgets when scale is small and integer rounding creates
    duplicates
    """
    assert compute_budgets(1, 16, 2, 10, 1) == [
        [
            (512, 1),
            (256, 2),
            (128, 3),
            (64, 4),
            (32, 5),
            (16, 6),
            (8, 7),
            (4, 9),
            (2, 12),
            (1, 16),
        ]
    ]

    with pytest.raises(ValueError) as exc:
        compute_budgets(1, 2, 2, 10, 1)

    assert "Cannot build budgets below max_resources" in str(exc.value)


class TestASHABracket:
    """Tests for the `ASHABracket` class."""

    def test_rungs_creation(self, bracket):
        """Test the creation of rungs for bracket 0."""
        assert len(bracket.rungs) == 3
        assert bracket.rungs[0]["resources"] == 1
        assert bracket.rungs[1]["resources"] == 3
        assert bracket.rungs[2]["resources"] == 9

    def test_register(self, asha, bracket):
        """Check that a point is correctly registered inside a bracket."""
        bracket.asha = asha
        point = (1, 0.0)
        point_hash = hashlib.md5(str([0.0]).encode("utf-8")).hexdigest()

        bracket.register(point, 0.0)

        assert len(bracket.rungs[0])
        assert point_hash in bracket.rungs[0]["results"]
        assert (0.0, point) == bracket.rungs[0]["results"][point_hash]

    def test_bad_register(self, asha, bracket):
        """Check that a non-valid point is not registered."""
        bracket.asha = asha

        with pytest.raises(IndexError) as ex:
            bracket.register((55, 0.0), 0.0)

        assert "Bad fidelity level 55" in str(ex.value)

    def test_candidate_promotion(self, asha, bracket, rung_0):
        """Test that correct point is promoted."""
        bracket.asha = asha
        bracket.rungs[0] = rung_0

        point = bracket.get_candidate(0)

        assert point == (1, 0.0)

    def test_promotion_with_rung_1_hit(self, asha, bracket, rung_0):
        """Test that get_candidate gives us the next best thing if point is already in rung 1."""
        point = (1, 0.0)
        point_hash = hashlib.md5(str([0.0]).encode("utf-8")).hexdigest()
        bracket.asha = asha
        bracket.rungs[0] = rung_0
        bracket.rungs[1]["results"][point_hash] = (0.0, point)

        point = bracket.get_candidate(0)

        assert point == (1, 0.125)

    def test_no_promotion_when_rung_full(self, asha, bracket, rung_0, rung_1):
        """Test that get_candidate returns `None` if rung 1 is full."""
        bracket.asha = asha
        bracket.rungs[0] = rung_0
        bracket.rungs[1] = rung_1

        point = bracket.get_candidate(0)

        assert point is None

    def test_no_promotion_if_not_enough_points(self, asha, bracket):
        """Test the get_candidate return None if there is not enough points ready."""
        bracket.asha = asha
        bracket.rungs[0] = dict(
            n_trials=1,
            resources=1,
            results={
                hashlib.md5(str([0.0]).encode("utf-8")).hexdigest(): (0.0, (1, 0.0))
            },
        )

        point = bracket.get_candidate(0)

        assert point is None

    def test_no_promotion_if_not_completed(self, asha, bracket, rung_0):
        """Test the get_candidate return None if trials are not completed."""
        bracket.asha = asha
        bracket.rungs[0] = rung_0
        rung = bracket.rungs[0]["results"]

        point = bracket.get_candidate(0)

        for p_id in rung.keys():
            rung[p_id] = (None, rung[p_id][1])

        point = bracket.get_candidate(0)

        assert point is None

    def test_is_done(self, bracket, rung_0):
        """Test that the `is_done` property works."""
        assert not bracket.is_done

        # Actual value of the point is not important here
        bracket.rungs[2] = dict(n_trials=1, resources=9, results={"1": (1, 0.0)})

        assert bracket.is_done

    def test_update_rungs_return_candidate(self, asha, bracket, rung_1):
        """Check if a valid modified candidate is returned by update_rungs."""
        bracket.asha = asha
        bracket.rungs[1] = rung_1
        point_hash = hashlib.md5(str([0.0]).encode("utf-8")).hexdigest()

        candidate = bracket.promote(1)[0]

        assert point_hash in bracket.rungs[1]["results"]
        assert bracket.rungs[1]["results"][point_hash] == (0.0, (3, 0.0))
        assert candidate[0] == 9

    def test_update_rungs_return_no_candidate(self, asha, bracket, rung_1):
        """Check if no candidate is returned by update_rungs."""
        bracket.asha = asha

        candidate = bracket.promote(1)

        assert candidate == []

    def test_repr(self, bracket, rung_0, rung_1, rung_2):
        """Test the string representation of ASHABracket"""
        bracket.rungs[0] = rung_0
        bracket.rungs[1] = rung_1
        bracket.rungs[2] = rung_2

        assert str(bracket) == "ASHABracket(resource=[1, 3, 9], repetition id=1)"


class TestASHA:
    """Tests for the algo ASHA."""

    def test_register(self, asha, bracket, rung_0, rung_1):
        """Check that a point is registered inside the bracket."""
        asha.brackets = [bracket]
        bracket.asha = asha
        bracket.rungs = [rung_0, rung_1]
        point = (1, 0.0)
        point_hash = hashlib.md5(str([0.0]).encode("utf-8")).hexdigest()

        asha.observe([point], [{"objective": 0.0}])

        assert len(bracket.rungs[0])
        assert point_hash in bracket.rungs[0]["results"]
        assert (0.0, point) == bracket.rungs[0]["results"][point_hash]

    def test_register_bracket_multi_fidelity(self, space, b_config):
        """Check that a point is registered inside the same bracket for diff fidelity."""
        asha = ASHA(space, num_brackets=3)

        value = 50
        fidelity = 1
        point = (fidelity, value)
        point_hash = hashlib.md5(str([value]).encode("utf-8")).hexdigest()

        force_observe(asha, point, {"objective": 0.0})

        bracket = asha.brackets[0]

        assert len(bracket.rungs[0])
        assert point_hash in bracket.rungs[0]["results"]
        assert (0.0, point) == bracket.rungs[0]["results"][point_hash]

        fidelity = 3
        point = [fidelity, value]
        point_hash = hashlib.md5(str([value]).encode("utf-8")).hexdigest()

        force_observe(asha, point, {"objective": 0.0})

        assert len(bracket.rungs[0])
        assert point_hash in bracket.rungs[1]["results"]
        assert (0.0, point) != bracket.rungs[0]["results"][point_hash]
        assert (0.0, point) == bracket.rungs[1]["results"][point_hash]

    def test_register_next_bracket(self, space, b_config):
        """Check that a point is registered inside the good bracket when higher fidelity."""
        asha = ASHA(space, num_brackets=3)

        value = 50
        fidelity = 3
        point = (fidelity, value)
        point_hash = hashlib.md5(str([value]).encode("utf-8")).hexdigest()

        force_observe(asha, point, {"objective": 0.0})

        assert sum(len(rung["results"]) for rung in asha.brackets[0].rungs) == 0
        assert sum(len(rung["results"]) for rung in asha.brackets[1].rungs) == 1
        assert sum(len(rung["results"]) for rung in asha.brackets[2].rungs) == 0
        assert point_hash in asha.brackets[1].rungs[0]["results"]
        assert (0.0, point) == asha.brackets[1].rungs[0]["results"][point_hash]

        value = 51
        fidelity = 9
        point = (fidelity, value)
        point_hash = hashlib.md5(str([value]).encode("utf-8")).hexdigest()

        force_observe(asha, point, {"objective": 0.0})

        assert sum(len(rung["results"]) for rung in asha.brackets[0].rungs) == 0
        assert sum(len(rung["results"]) for rung in asha.brackets[1].rungs) == 1
        assert sum(len(rung["results"]) for rung in asha.brackets[2].rungs) == 1
        assert point_hash in asha.brackets[2].rungs[0]["results"]
        assert (0.0, point) == asha.brackets[2].rungs[0]["results"][point_hash]

    def test_register_invalid_fidelity(self, space, b_config):
        """Check that a point cannot registered if fidelity is invalid."""
        asha = ASHA(space, num_brackets=3)

        value = 50
        fidelity = 2
        point = (fidelity, value)

        with pytest.raises(ValueError) as ex:
            force_observe(asha, point, {"objective": 0.0})

        assert "No bracket found for point" in str(ex.value)

    def test_register_not_sampled(self, space, b_config, caplog):
        """Check that a point cannot registered if not sampled."""
        asha = ASHA(space, num_brackets=3)

        value = 50
        fidelity = 2
        point = (fidelity, value)

        with caplog.at_level(logging.INFO, logger="orion.algo.hyperband"):
            asha.observe([point], [{"objective": 0.0}])

        assert len(caplog.records) == 1
        assert "Ignoring point" in caplog.records[0].msg

    def test_register_corrupted_db(self, caplog, space, b_config):
        """Check that a point cannot registered if passed in order diff than fidelity."""
        asha = ASHA(space, num_brackets=3)

        value = 50
        fidelity = 3
        point = (fidelity, value)

        force_observe(asha, point, {"objective": 0.0})
        assert "Point registered to wrong bracket" not in caplog.text

        fidelity = 1
        point = [fidelity, value]

        caplog.clear()
        force_observe(asha, point, {"objective": 0.0})
        assert "Point registered to wrong bracket" in caplog.text

    def test_suggest_new(self, monkeypatch, asha, bracket, rung_0, rung_1, rung_2):
        """Test that a new point is sampled."""
        asha.brackets = [bracket]
        bracket.asha = asha

        def sample(num=1, seed=None):
            return [("fidelity", 0.5)]

        monkeypatch.setattr(asha.space, "sample", sample)

        points = asha.suggest(1)

        assert points == [(1, 0.5)]

    def test_suggest_duplicates(
        self, monkeypatch, asha, bracket, rung_0, rung_1, rung_2
    ):
        """Test that sampling collisions are handled."""
        asha.brackets = [bracket]
        bracket.asha = asha

        duplicate_point = ("fidelity", 0.0)
        new_point = ("fidelity", 0.5)

        duplicate_id_wo_fidelity = asha.get_id(duplicate_point, ignore_fidelity=True)
        bracket.rungs[0] = dict(
            n_trials=2,
            resources=1,
            results={duplicate_id_wo_fidelity: (0.0, duplicate_point)},
        )
        asha.trial_to_brackets[duplicate_id_wo_fidelity] = bracket

        asha.register(duplicate_point, 0.0)

        points = [duplicate_point, new_point]

        def sample(num=1, seed=None):
            return points

        monkeypatch.setattr(asha.space, "sample", sample)

        assert asha.suggest(1)[0][1] == new_point[1]

    def test_suggest_inf_duplicates(
        self, monkeypatch, asha, bracket, rung_0, rung_1, rung_2
    ):
        """Test that sampling inf collisions returns None."""
        asha.brackets = [bracket]
        bracket.asha = asha

        zhe_point = ("fidelity", 0.0)
        asha.trial_to_brackets[asha.get_id(zhe_point, ignore_fidelity=True)] = bracket

        def sample(num=1, seed=None):
            return [zhe_point]

        monkeypatch.setattr(asha.space, "sample", sample)

        assert asha.suggest(1) == []

    def test_suggest_in_finite_cardinality(self):
        """Test that suggest None when search space is empty"""
        space = Space()
        space.register(Integer("yolo1", "uniform", 0, 5))
        space.register(Fidelity("epoch", 1, 9, 3))

        asha = ASHA(space)
        for i in range(6):
            force_observe(asha, (1, i), {"objective": i})

        for i in range(2):
            force_observe(asha, (3, i), {"objective": i})

        assert asha.suggest(1) == []

    def test_suggest_promote(self, asha, bracket, rung_0):
        """Test that correct point is promoted and returned."""
        asha.brackets = [bracket]
        bracket.asha = asha
        bracket.rungs[0] = rung_0

        points = asha.suggest(1)

        assert points == [(3, 0.0)]


class TestGenericASHA(BaseAlgoTests):
    algo_name = "asha"
    config = {
        "seed": 123456,
        "num_rungs": 5,
        "num_brackets": 2,
        "repetitions": 3,
    }
    space = {"x": "uniform(0, 1)", "f": "fidelity(1, 10, base=2)"}

    def test_suggest_n(self, mocker, num, attr):
        algo = self.create_algo()
        spy = self.spy_phase(mocker, num, algo, attr)
        points = algo.suggest(5)
        assert len(points) == 1

    @pytest.mark.skip(reason="See https://github.com/Epistimio/orion/issues/598")
    def test_is_done_cardinality(self):
        space = self.update_space(
            {
                "x": "uniform(0, 4, discrete=True)",
                "y": "choices(['a', 'b', 'c'])",
                "z": "loguniform(1, 6, discrete=True)",
            }
        )
        space = self.create_space(space)
        assert space.cardinality == 5 * 3 * 6

        algo = self.create_algo(num_brackets=2, space=space)

        for rung in range(algo.algorithm.num_rungs):
            assert not algo.is_done

            points = []
            while True:
                assert not algo.is_done
                n_sampled = len(algo.algorithm.sampled)
                n_trials = len(algo.algorithm.trial_to_brackets)
                new_points = algo.suggest(1)
                if new_points is None:
                    break
                points += new_points
                if rung == 0:
                    assert len(algo.algorithm.sampled) == n_sampled + 1
                else:
                    assert len(algo.algorithm.sampled) == n_sampled
                assert len(algo.algorithm.trial_to_brackets) == n_trials + 1

            assert not algo.is_done

            for i, point in enumerate(points):
                algo.observe([point], [dict(objective=i)])

        assert algo.is_done

    def test_is_done_max_trials(self):
        space = self.create_space()

        MAX_TRIALS = 10
        algo = self.create_algo(space=space)
        algo.algorithm.max_trials = MAX_TRIALS

        objective = 0
        while not algo.is_done:
            points = algo.suggest(1)
            assert points is not None
            if points:
                self.observe_points(points, algo, objective)
                objective += len(points)

        # ASHA should ignore max trials.
        assert algo.n_observed > MAX_TRIALS
        assert algo.is_done

    @pytest.mark.skip(reason="See https://github.com/Epistimio/orion/issues/599")
    def test_optimize_branin(self):
        pass

    def infer_repetition_and_rung(self, num):
        budgets = BUDGETS
        if num > budgets[-1] * 2:
            return 3, 0
        elif num > budgets[-1]:
            return 2, 0

        if num == 0:
            return 1, -1

        return 1, budgets.index(num)

    def assert_callbacks(self, spy, num, algo):

        if num == 0:
            return

        repetition_id, rung_id = self.infer_repetition_and_rung(num)

        brackets = algo.algorithm.brackets

        assert len(brackets) == repetition_id * 2

        for j in range(0, rung_id + 1):
            for bracket in brackets:
                assert len(bracket.rungs[j]["results"]) > 0, (bracket, j)


BUDGETS = [
    4 + 2,  # rung 0
    (8 + 4 + 2 + 4),  # rung 1 (first bracket 8 4 2, second bracket 4)
    (16 + 8 + 4 + 2 + 1 + 8 + 4 + 2),  #  rung 2
]

TestGenericASHA.set_phases(
    [("random", 0, "space.sample")]
    + [(f"rung{i}", budget, "suggest") for i, budget in enumerate(BUDGETS)]
    + [("rep1-rung1", BUDGETS[-1] + 16, "suggest")]
    + [("rep2-rung1", BUDGETS[-1] * 2 + 16, "suggest")]
)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Collection of tests for :mod:`orion.storage`."""

import copy
import datetime
import logging
import os
import pickle
import time

import pytest

import orion.core
from orion.core.io.database import DuplicateKeyError
from orion.core.io.database.pickleddb import PickledDB
from orion.core.utils.singleton import (
    SingletonAlreadyInstantiatedError,
    SingletonNotInstantiatedError,
    update_singletons,
)
from orion.core.worker.trial import Trial
from orion.storage.base import (
    FailedUpdate,
    MissingArguments,
    Storage,
    get_storage,
    setup_storage,
)
from orion.storage.legacy import Legacy
from orion.storage.track import HAS_TRACK, REASON
from orion.testing import OrionState

log = logging.getLogger(__name__)
log.setLevel(logging.WARNING)

storage_backends = [None]  # defaults to legacy with PickleDB

if not HAS_TRACK:
    log.warning("Track is not tested because: %s!", REASON)
else:
    storage_backends.append({"type": "track", "uri": "file://${file}?objective=loss"})

base_experiment = {
    "name": "default_name",
    "version": 0,
    "metadata": {
        "user": "default_user",
        "user_script": "abc",
        "priors": {"x": "uniform(0, 10)"},
        "datetime": "2017-11-23T02:00:00",
        "orion_version": "XYZ",
    },
}


base_trial = {
    "experiment": "default_name",
    "status": "new",  # new, reserved, suspended, completed, broken
    "worker": None,
    "submit_time": "2017-11-23T02:00:00",
    "start_time": None,
    "end_time": None,
    "heartbeat": None,
    "results": [
        {"name": "loss", "type": "objective", "value": 2}  # objective, constraint
    ],
    "params": [
        {"name": "/encoding_layer", "type": "categorical", "value": "rnn"},
        {
            "name": "/decoding_layer",
            "type": "categorical",
            "value": "lstm_with_attention",
        },
    ],
}


def _generate(obj, *args, value):
    if obj is None:
        return None

    obj = copy.deepcopy(obj)
    data = obj

    for arg in args[:-1]:
        data = data[arg]

    data[args[-1]] = value
    return obj


def make_lost_trial(delay=2):
    """Make a lost trial"""
    obj = copy.deepcopy(base_trial)
    obj["status"] = "reserved"
    obj["heartbeat"] = datetime.datetime.utcnow() - datetime.timedelta(
        seconds=orion.core.config.worker.heartbeat * delay
    )
    obj["params"].append(
        {"name": "/index", "type": "categorical", "value": f"lost_trial_{delay}"}
    )
    return obj


all_status = ["completed", "broken", "reserved", "interrupted", "suspended", "new"]


def generate_trials(status=None, heartbeat=None):
    """Generate Trials with different configurations"""
    if status is None:
        status = all_status

    new_trials = [_generate(base_trial, "status", value=s) for s in status]

    if heartbeat:
        for trial in new_trials:
            trial["heartbeat"] = heartbeat

    # make each trial unique
    for i, trial in enumerate(new_trials):
        trial["params"].append({"name": "/index", "type": "categorical", "value": i})

    return new_trials


def generate_experiments():
    """Generate a set of experiments"""
    users = ["a", "b", "c"]
    exps = [_generate(base_experiment, "metadata", "user", value=u) for u in users]
    return [_generate(exp, "name", value=str(i)) for i, exp in enumerate(exps)]


@pytest.mark.usefixtures("setup_pickleddb_database")
def test_setup_storage_default():
    """Test that storage is setup using default config"""
    update_singletons()
    setup_storage()
    storage = Storage()
    assert isinstance(storage, Legacy)
    assert isinstance(storage._db, PickledDB)


def test_setup_storage_bad():
    """Test how setup fails when configuring with non-existant backends"""
    update_singletons()
    with pytest.raises(NotImplementedError) as exc:
        setup_storage({"type": "idontexist"})

    assert exc.match("idontexist")


def test_setup_storage_custom():
    """Test setup with local configuration"""
    update_singletons()
    setup_storage(
        {"type": "legacy", "database": {"type": "pickleddb", "host": "test.pkl"}}
    )
    storage = Storage()
    assert isinstance(storage, Legacy)
    assert isinstance(storage._db, PickledDB)
    assert storage._db.host == os.path.abspath("test.pkl")


def test_setup_storage_custom_type_missing():
    """Test setup with local configuration with type missing"""
    update_singletons()
    setup_storage({"database": {"type": "pickleddb", "host": "test.pkl"}})
    storage = Storage()
    assert isinstance(storage, Legacy)
    assert isinstance(storage._db, PickledDB)
    assert storage._db.host == os.path.abspath("test.pkl")


@pytest.mark.usefixtures("setup_pickleddb_database")
def test_setup_storage_custom_legacy_emtpy():
    """Test setup with local configuration with legacy but no config"""
    update_singletons()
    setup_storage({"type": "legacy"})
    storage = Storage()
    assert isinstance(storage, Legacy)
    assert isinstance(storage._db, PickledDB)
    assert storage._db.host == orion.core.config.storage.database.host


def test_setup_storage_bad_override():
    """Test setup with different type than existing singleton"""
    update_singletons()
    setup_storage(
        {"type": "legacy", "database": {"type": "pickleddb", "host": "test.pkl"}}
    )
    storage = Storage()
    assert isinstance(storage, Legacy)
    assert isinstance(storage._db, PickledDB)
    with pytest.raises(SingletonAlreadyInstantiatedError) as exc:
        setup_storage({"type": "track"})

    assert exc.match("A singleton instance of \(type: Storage\)")


@pytest.mark.xfail(reason="Fix this when introducing #135 in v0.2.0")
def test_setup_storage_bad_config_override():
    """Test setup with different config than existing singleton"""
    update_singletons()
    setup_storage({"database": {"type": "pickleddb", "host": "test.pkl"}})
    storage = Storage()
    assert isinstance(storage, Legacy)
    assert isinstance(storage._db, PickledDB)
    with pytest.raises(SingletonAlreadyInstantiatedError):
        setup_storage({"database": {"type": "mongodb"}})


def test_setup_storage_stateless():
    """Test that passed configuration dictionary is not modified by the fonction"""
    update_singletons()
    config = {"database": {"type": "pickleddb", "host": "test.pkl"}}
    passed_config = copy.deepcopy(config)
    setup_storage(passed_config)
    assert config == passed_config


def test_get_storage_uninitiated():
    """Test that get storage fails if no storage singleton exist"""
    update_singletons()
    with pytest.raises(SingletonNotInstantiatedError) as exc:
        get_storage()

    assert exc.match("No singleton instance of \(type: Storage\) was created")


def test_get_storage():
    """Test that get storage gets the singleton"""
    update_singletons()
    setup_storage({"database": {"type": "pickleddb", "host": "test.pkl"}})
    storage = get_storage()
    assert isinstance(storage, Legacy)
    assert isinstance(storage._db, PickledDB)
    assert get_storage() == storage


@pytest.mark.usefixtures("version_XYZ")
@pytest.mark.parametrize("storage", storage_backends)
class TestStorage:
    """Test all storage backend"""

    def test_create_experiment(self, storage):
        """Test create experiment"""
        with OrionState(experiments=[], storage=storage) as cfg:
            storage = cfg.storage()

            storage.create_experiment(base_experiment)

            experiments = storage.fetch_experiments({})
            assert len(experiments) == 1, "Only one experiment in the database"

            experiment = experiments[0]
            assert base_experiment == experiment, "Local experiment and DB should match"

            # Insert it again
            with pytest.raises(DuplicateKeyError):
                storage.create_experiment(base_experiment)

    def test_fetch_experiments(self, storage, name="0", user="a"):
        """Test fetch experiments"""
        with OrionState(experiments=generate_experiments(), storage=storage) as cfg:
            storage = cfg.storage()

            experiments = storage.fetch_experiments({})
            assert len(experiments) == len(cfg.experiments)

            experiments = storage.fetch_experiments(
                {"name": name, "metadata.user": user}
            )
            assert len(experiments) == 1

            experiment = experiments[0]
            assert experiment["name"] == name, "name should match query"
            assert (
                experiment["metadata"]["user"] == user
            ), "user name should match query"

            experiments = storage.fetch_experiments(
                {"name": "-1", "metadata.user": user}
            )
            assert len(experiments) == 0

    def test_update_experiment(self, monkeypatch, storage, name="0", user="a"):
        """Test fetch experiments"""
        with OrionState(experiments=generate_experiments(), storage=storage) as cfg:
            storage = cfg.storage()

            class _Dummy:
                pass

            experiment = cfg.experiments[0]
            mocked_experiment = _Dummy()
            mocked_experiment.id = experiment["_id"]

            storage.update_experiment(mocked_experiment, test=True)
            experiments = storage.fetch_experiments({"_id": experiment["_id"]})
            assert len(experiments) == 1

            fetched_experiment = experiments[0]
            assert fetched_experiment["test"]

            assert (
                "test"
                not in storage.fetch_experiments({"_id": cfg.experiments[1]["_id"]})[0]
            )

            storage.update_experiment(uid=experiment["_id"], test2=True)
            assert storage.fetch_experiments({"_id": experiment["_id"]})[0]["test2"]
            assert (
                "test2"
                not in storage.fetch_experiments({"_id": cfg.experiments[1]["_id"]})[0]
            )

            with pytest.raises(MissingArguments):
                storage.update_experiment()

            with pytest.raises(AssertionError):
                storage.update_experiment(experiment=mocked_experiment, uid="123")

    def test_delete_experiment(self, storage):
        """Test delete one experiment"""
        if storage and storage["type"] == "track":
            pytest.xfail("Track does not support deletion yet.")

        with OrionState(experiments=generate_experiments(), storage=storage) as cfg:
            storage = cfg.storage()

            n_experiments = len(storage.fetch_experiments({}))
            storage.delete_experiment(uid=cfg.experiments[0]["_id"])
            experiments = storage.fetch_experiments({})
            assert len(experiments) == n_experiments - 1
            assert cfg.experiments[0]["_id"] not in [exp["_id"] for exp in experiments]

    def test_register_trial(self, storage):
        """Test register trial"""
        with OrionState(experiments=[base_experiment], storage=storage) as cfg:
            storage = cfg.storage()
            trial1 = storage.register_trial(Trial(**base_trial))
            trial2 = storage.get_trial(trial1)

            assert (
                trial1.to_dict() == trial2.to_dict()
            ), "Trials should match after insert"

    def test_register_duplicate_trial(self, storage):
        """Test register trial"""
        with OrionState(
            experiments=[base_experiment], trials=[base_trial], storage=storage
        ) as cfg:
            storage = cfg.storage()

            with pytest.raises(DuplicateKeyError):
                storage.register_trial(Trial(**base_trial))

    def test_register_lie(self, storage):
        """Test register lie"""
        with OrionState(experiments=[base_experiment], storage=storage) as cfg:
            storage = cfg.storage()
            storage.register_lie(Trial(**base_trial))

    def test_register_lie_fail(self, storage):
        """Test register lie"""
        with OrionState(
            experiments=[base_experiment], lies=[base_trial], storage=storage
        ) as cfg:
            storage = cfg.storage()

            with pytest.raises(DuplicateKeyError):
                storage.register_lie(Trial(**cfg.lies[0]))

    def test_update_trials(self, storage):
        """Test update many trials"""
        with OrionState(
            experiments=[base_experiment],
            trials=generate_trials(status=["completed", "reserved", "reserved"]),
            storage=storage,
        ) as cfg:
            storage = cfg.storage()

            class _Dummy:
                pass

            experiment = cfg.get_experiment("default_name", version=None)
            trials = storage.fetch_trials(experiment)
            assert sum(trial.status == "reserved" for trial in trials) == 2
            count = storage.update_trials(
                experiment, where={"status": "reserved"}, status="interrupted"
            )
            assert count == 2
            trials = storage.fetch_trials(experiment)
            assert sum(trial.status == "interrupted" for trial in trials) == 2

    def test_update_trial(self, storage):
        """Test update one trial"""
        with OrionState(
            experiments=[base_experiment], trials=[base_trial], storage=storage
        ) as cfg:
            storage = cfg.storage()

            trial = Trial(**cfg.trials[0])

            assert trial.status != "interrupted"
            storage.update_trial(trial, status="interrupted")
            assert storage.get_trial(trial).status == "interrupted"

    def test_reserve_trial_success(self, storage):
        """Test reserve trial"""
        with OrionState(
            experiments=[base_experiment], trials=[base_trial], storage=storage
        ) as cfg:
            storage = cfg.storage()
            experiment = cfg.get_experiment("default_name", version=None)

            trial = storage.reserve_trial(experiment)

            assert trial is not None
            assert trial.status == "reserved"

    def test_reserve_trial_fail(self, storage):
        """Test reserve trial"""
        with OrionState(
            experiments=[base_experiment],
            trials=generate_trials(status=["completed", "reserved"]),
            storage=storage,
        ) as cfg:

            storage = cfg.storage()
            experiment = cfg.get_experiment("default_name", version=None)

            trial = storage.reserve_trial(experiment)
            assert trial is None

    def test_fetch_trials(self, storage):
        """Test fetch experiment trials"""
        with OrionState(
            experiments=[base_experiment], trials=generate_trials(), storage=storage
        ) as cfg:
            storage = cfg.storage()
            experiment = cfg.get_experiment("default_name", version=None)

            trials1 = storage.fetch_trials(experiment=experiment)
            trials2 = storage.fetch_trials(uid=experiment.id)

            with pytest.raises(MissingArguments):
                storage.fetch_trials()

            with pytest.raises(AssertionError):
                storage.fetch_trials(experiment=experiment, uid="123")

            assert len(trials1) == len(cfg.trials), "trial count should match"
            assert len(trials2) == len(cfg.trials), "trial count should match"

    def test_fetch_trials_with_query(self, storage):
        """Test fetch experiment trials with queries"""
        with OrionState(
            experiments=[base_experiment],
            trials=generate_trials(status=["completed", "reserved", "reserved"]),
            storage=storage,
        ) as cfg:
            storage = cfg.storage()
            experiment = cfg.get_experiment("default_name", version=None)

            trials_all = storage.fetch_trials(experiment=experiment)
            trials_completed = storage.fetch_trials(
                experiment=experiment, where={"status": "completed"}
            )
            trials_reserved = storage.fetch_trials(
                experiment=experiment, where={"status": "reserved"}
            )

            assert len(trials_all) == len(cfg.trials), "trial count should match"
            assert len(trials_completed) == 1, "trial count should match"
            assert len(trials_reserved) == 2, "trial count should match"

    def test_delete_all_trials(self, storage):
        """Test delete all trials of an experiment"""
        if storage and storage["type"] == "track":
            pytest.xfail("Track does not support deletion yet.")

        trials = generate_trials()
        trial_from_other_exp = copy.deepcopy(trials[0])
        trial_from_other_exp["experiment"] = "other"
        trials.append(trial_from_other_exp)
        with OrionState(
            experiments=[base_experiment], trials=trials, storage=storage
        ) as cfg:
            storage = cfg.storage()

            # Make sure we have sufficient trials to test deletion
            trials = storage.fetch_trials(uid="default_name")
            assert len(trials) > 2

            count = storage.delete_trials(uid="default_name")
            assert count == len(trials)
            assert storage.fetch_trials(uid="default_name") == []

            # Make sure trials from other experiments were not deleted
            assert len(storage.fetch_trials(uid="other")) == 1

    def test_delete_trials_with_query(self, storage):
        """Test delete experiment trials matching a query"""
        if storage and storage["type"] == "track":
            pytest.xfail("Track does not support deletion yet.")

        trials = generate_trials()
        trial_from_other_exp = copy.deepcopy(trials[0])
        trial_from_other_exp["experiment"] = "other"
        trials.append(trial_from_other_exp)
        with OrionState(
            experiments=[base_experiment], trials=trials, storage=storage
        ) as cfg:
            storage = cfg.storage()
            experiment = cfg.get_experiment("default_name")

            # Make sure we have sufficient trials to test deletion
            status = trials[0]["status"]
            trials = storage.fetch_trials(experiment)
            trials_with_status = storage.fetch_trials_by_status(experiment, status)
            assert len(trials_with_status) > 0
            assert len(trials) > len(trials_with_status)

            # Test deletion
            count = storage.delete_trials(uid="default_name", where={"status": status})
            assert count == len(trials_with_status)
            assert storage.fetch_trials_by_status(experiment, status) == []
            assert len(storage.fetch_trials(experiment)) == len(trials) - len(
                trials_with_status
            )

            # Make sure trials from other experiments were not deleted
            assert len(storage.fetch_trials(uid="other")) == 1

    def test_get_trial(self, storage):
        """Test get trial"""
        with OrionState(
            experiments=[base_experiment], trials=generate_trials(), storage=storage
        ) as cfg:
            storage = cfg.storage()

            trial_dict = cfg.trials[0]

            trial1 = storage.get_trial(trial=Trial(**trial_dict))
            trial2 = storage.get_trial(uid=trial1.id)

            with pytest.raises(MissingArguments):
                storage.get_trial()

            with pytest.raises(AssertionError):
                storage.get_trial(trial=trial1, uid="123")

            assert trial1.to_dict() == trial_dict
            assert trial2.to_dict() == trial_dict

    def test_fetch_lost_trials(self, storage):
        """Test update heartbeat"""
        lost_trials = [
            make_lost_trial(2),  # Is not lost for long enough to be catched
            make_lost_trial(10),
        ]  # Is lost for long enough to be catched
        # Force recent heartbeat to avoid mixing up with lost trials.
        trials = lost_trials + generate_trials(heartbeat=datetime.datetime.utcnow())
        with OrionState(
            experiments=[base_experiment], trials=trials, storage=storage
        ) as cfg:
            storage = cfg.storage()

            experiment = cfg.get_experiment("default_name", version=None)
            trials = storage.fetch_lost_trials(experiment)

            assert len(trials) == 1

    def test_change_status_success(self, storage):
        """Change the status of a Trial"""

        def check_status_change(new_status):
            with OrionState(
                experiments=[base_experiment], trials=generate_trials(), storage=storage
            ) as cfg:
                trial = get_storage().get_trial(cfg.get_trial(0))
                assert trial is not None, "was not able to retrieve trial for test"

                get_storage().set_trial_status(trial, status=new_status)
                assert (
                    trial.status == new_status
                ), "Trial status should have been updated locally"

                trial = get_storage().get_trial(trial)
                assert (
                    trial.status == new_status
                ), "Trial status should have been updated in the storage"

        check_status_change("completed")
        check_status_change("broken")
        check_status_change("reserved")
        check_status_change("interrupted")
        check_status_change("suspended")
        check_status_change("new")

    def test_change_status_invalid(self, storage):
        """Attempt to change the status of a Trial with an invalid one"""
        with OrionState(
            experiments=[base_experiment], trials=generate_trials(), storage=storage
        ) as cfg:
            trial = get_storage().get_trial(cfg.get_trial(0))
            assert trial is not None, "Was not able to retrieve trial for test"

            with pytest.raises(ValueError) as exc:
                get_storage().set_trial_status(trial, status="moo")

            assert exc.match("Given status `moo` not one of")

    def test_change_status_failed_update(self, storage):
        """Change the status of a Trial"""

        def check_status_change(new_status):
            with OrionState(
                experiments=[base_experiment], trials=generate_trials(), storage=storage
            ) as cfg:
                trial = get_storage().get_trial(cfg.get_trial(0))
                assert trial is not None, "Was not able to retrieve trial for test"
                assert trial.status != new_status

                if trial.status == new_status:
                    return

                with pytest.raises(FailedUpdate):
                    trial.status = new_status
                    get_storage().set_trial_status(trial, status=new_status)

        check_status_change("completed")
        check_status_change("broken")
        check_status_change("reserved")
        check_status_change("interrupted")
        check_status_change("suspended")

    def test_change_status_success_thanks_to_was(self, storage):
        """Change the status of a Trial requesting the correct previous state, although local
        one is not up-to-date.
        """

        def check_status_change(new_status):
            with OrionState(
                experiments=[base_experiment], trials=generate_trials(), storage=storage
            ) as cfg:
                trial = get_storage().get_trial(cfg.get_trial(0))
                assert trial is not None, "Was not able to retrieve trial for test"
                assert trial.status != new_status

                if trial.status == new_status:
                    return

                correct_status = trial.status
                trial.status = "broken"
                assert correct_status != "broken"
                with pytest.raises(FailedUpdate):
                    get_storage().set_trial_status(trial, status=new_status)

                get_storage().set_trial_status(
                    trial, status=new_status, was=correct_status
                )

        check_status_change("completed")
        check_status_change("broken")
        check_status_change("reserved")
        check_status_change("interrupted")
        check_status_change("suspended")

    def test_change_status_failed_update_because_of_was(self, storage):
        """Change the status of a Trial requesting the wrong previous state."""

        def check_status_change(new_status):
            with OrionState(
                experiments=[base_experiment], trials=generate_trials(), storage=storage
            ) as cfg:
                trial = get_storage().get_trial(cfg.get_trial(0))
                assert trial is not None, "Was not able to retrieve trial for test"
                assert trial.status != new_status

                if trial.status == new_status:
                    return

                with pytest.raises(FailedUpdate):
                    get_storage().set_trial_status(
                        trial, status=new_status, was=new_status
                    )

        check_status_change("completed")
        check_status_change("broken")
        check_status_change("reserved")
        check_status_change("interrupted")
        check_status_change("suspended")

    def test_fetch_pending_trials(self, storage):
        """Test fetch pending trials"""
        with OrionState(
            experiments=[base_experiment], trials=generate_trials(), storage=storage
        ) as cfg:
            storage = cfg.storage()

            experiment = cfg.get_experiment("default_name", version=None)
            trials = storage.fetch_pending_trials(experiment)

            count = 0
            for trial in cfg.trials:
                if trial["status"] in {"new", "suspended", "interrupted"}:
                    count += 1

            assert len(trials) == count
            for trial in trials:
                assert trial.status in {"new", "suspended", "interrupted"}

    def test_fetch_noncompleted_trials(self, storage):
        """Test fetch non completed trials"""
        with OrionState(
            experiments=[base_experiment], trials=generate_trials(), storage=storage
        ) as cfg:
            storage = cfg.storage()

            experiment = cfg.get_experiment("default_name", version=None)
            trials = storage.fetch_noncompleted_trials(experiment)

            count = 0
            for trial in cfg.trials:
                if trial["status"] != "completed":
                    count += 1

            for trial in trials:
                assert trial.status != "completed"

            assert len(trials) == count

    def test_fetch_trials_by_status(self, storage):
        """Test fetch completed trials"""
        with OrionState(
            experiments=[base_experiment], trials=generate_trials(), storage=storage
        ) as cfg:
            count = 0
            for trial in cfg.trials:
                if trial["status"] == "completed":
                    count += 1

            storage = cfg.storage()
            experiment = cfg.get_experiment("default_name", version=None)
            trials = storage.fetch_trials_by_status(experiment, "completed")

            assert len(trials) == count
            for trial in trials:
                assert trial.status == "completed", trial

    def test_count_completed_trials(self, storage):
        """Test count completed trials"""
        with OrionState(
            experiments=[base_experiment], trials=generate_trials(), storage=storage
        ) as cfg:
            count = 0
            for trial in cfg.trials:
                if trial["status"] == "completed":
                    count += 1

            storage = cfg.storage()

            experiment = cfg.get_experiment("default_name", version=None)
            trials = storage.count_completed_trials(experiment)
            assert trials == count

    def test_count_broken_trials(self, storage):
        """Test count broken trials"""
        with OrionState(
            experiments=[base_experiment], trials=generate_trials(), storage=storage
        ) as cfg:
            count = 0
            for trial in cfg.trials:
                if trial["status"] == "broken":
                    count += 1

            storage = cfg.storage()

            experiment = cfg.get_experiment("default_name", version=None)

            trials = storage.count_broken_trials(experiment)

            assert trials == count

    def test_update_heartbeat(self, storage):
        """Test update heartbeat"""
        with OrionState(
            experiments=[base_experiment], trials=generate_trials(), storage=storage
        ) as cfg:
            storage_name = storage
            storage = cfg.storage()

            exp = cfg.get_experiment("default_name")
            trial1 = storage.fetch_trials_by_status(exp, status="reserved")[0]
            trial1b = copy.deepcopy(trial1)

            storage.update_heartbeat(trial1)

            trial2 = storage.get_trial(trial1)

            # this check that heartbeat is the correct type and that it was updated prior to now
            assert trial1b.heartbeat is None
            assert trial1.heartbeat is None
            assert trial2.heartbeat is not None

            # Sleep a bit, because fast CPUs make this test fail
            time.sleep(0.1)
            assert trial2.heartbeat < datetime.datetime.utcnow()

            if storage_name is None:
                trial3 = storage.fetch_trials_by_status(exp, status="completed")[0]
                storage.update_heartbeat(trial3)

                assert (
                    trial3.heartbeat is None
                ), "Legacy does not update trials with a status different from reserved"

    def test_serializable(self, storage):
        """Test storage can be serialized"""
        with OrionState(
            experiments=[base_experiment], trials=generate_trials(), storage=storage
        ) as cfg:
            storage = cfg.storage()
            serialized = pickle.dumps(storage)
            deserialized = pickle.loads(serialized)
            assert storage.fetch_experiments({}) == deserialized.fetch_experiments({})



"""
==============================
How to share with chart studio
==============================

This tutorial explains how to share plots using chart studio. This 
method that was used to share plots on the WordPress blog-post:
`Improved Deep Learning Workflows Through Hyperparameter Optimization with Oríon
<https://mila.quebec/en/article/improved-deep-learning-workflows-through-hyperparameter-optimization-with-orion/>`_.

Uploading to Chart Studio
-------------------------

If you do not already have a Chart Studio, you can create a 
`free one <https://plotly.com/api_signup>`_. 
With your account set up, you can get your `API key <https://plotly.com/settings/api>`_
set Chart Studio's credentials. 

"""
import chart_studio

username = "<your username>"
api_key = "<your api_key>"

chart_studio.tools.set_credentials_file(username=username, api_key=api_key)

#%%
# This will create a credential file locally that will be used in next step.
# Next you can create a ``figure`` and save it to Chart Studio.

# # Push your visualiztion to your account using the following lines of code:
from orion.client import get_experiment

# Specify the database where the experiments are stored. We use a local PickleDB here.
storage = dict(type="legacy", database=dict(type="pickleddb", host="../../db.pkl"))

fig = experiment.plot.regret()


import chart_studio.plotly as py

py.plot(fig, filename="regret")

# sphinx_gallery_thumbnail_path = '_static/share.png'

#%%
# The plot will be saved under the name ``regret`` in your
# `Chart Studio profile <https://chart-studio.plotly.com/organize/home>`_.
# Here is one example that was used in the blog-post mentioned earlier:
# `https://chart-studio.plotly.com/~xavier.bouthillier/1 <https://chart-studio.plotly.com/~xavier.bouthillier/1>`_.

#%%
# Sharing on WordPress
# --------------------
#
# Saving HTML version of plots
# does not work well with WordPress because it includes JavaScript.
# Using Chart Studio makes it possible to embed extarnal URL in WordPress posts.
#
# With Chart Studio,
# you can get an sharing URL for an embedded plot. In the Viewer page of Chart Studio,
# click on `export` and then `Embed URL`. This should bring you to a page with an URL
# formatted as ``https://chart-studio.plotly.com/~<your username>/<plot id>.embed``.
#
# With the `WordPress plugin iframe <https://fr.wordpress.org/plugins/iframe/>`_, you
# can embed your plot in a blog post. For example, the first plot in
# our blog-post is embedded with the following snippet.
#
# ::
#
#     [iframe src="//plotly.com/~xavier.bouthillier/1.embed?link=false&amp;autosize=true&amp;height=350"]
#
# Arguments to customized the embedded are documented
# `here
# <https://plotly.com/chart-studio-help/embed-graphs-in-websites/#step-8-customize-the-iframe>`_.
# See `starting guide <https://plotly.com/python/getting-started-with-chart-studio/>`_
# for more information on Chart Studio.



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Common fixtures for functional serving tests"""
import pytest
from falcon import testing

from orion.serving.webapi import WebApi
from orion.testing import OrionState


@pytest.fixture()
def client():
    """Mock the falcon.API instance for testing with an in memory database"""
    storage = {"type": "legacy", "database": {"type": "EphemeralDB"}}
    with OrionState(storage=storage):
        yield testing.TestClient(WebApi({"storage": storage}))



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Tests for :mod:`orion.benchmark.Benchmark`."""

import plotly
import pytest

import orion.core.io.experiment_builder as experiment_builder
from orion.benchmark import Benchmark, Study
from orion.benchmark.assessment import AverageRank, AverageResult
from orion.benchmark.task import CarromTable, RosenBrock
from orion.client.experiment import ExperimentClient
from orion.core.worker.experiment import Experiment
from orion.testing import OrionState, create_study_experiments


@pytest.fixture
def benchmark(benchmark_algorithms):
    """Return a benchmark instance"""
    return Benchmark(
        name="benchmark007",
        algorithms=benchmark_algorithms,
        targets=[
            {
                "assess": [AverageResult(2), AverageRank(2)],
                "task": [RosenBrock(25, dim=3), CarromTable(20)],
            }
        ],
    )


@pytest.fixture
def study(benchmark, benchmark_algorithms):
    """Return a study instance"""
    return Study(
        benchmark, benchmark_algorithms, AverageResult(2), RosenBrock(25, dim=3)
    )


class TestBenchmark:
    """Test Benchmark"""

    def test_creation(self, benchmark, benchmark_algorithms):
        """Test benchmark instance creation"""
        cfg = {
            "name": "benchmark007",
            "algorithms": benchmark_algorithms,
            "targets": [
                {
                    "assess": {
                        "AverageResult": {"task_num": 2},
                        "AverageRank": {"task_num": 2},
                    },
                    "task": {
                        "RosenBrock": {"dim": 3, "max_trials": 25},
                        "CarromTable": {"max_trials": 20},
                    },
                }
            ],
        }

        assert benchmark.configuration == cfg

    def test_setup_studies(self, benchmark):
        """Test to setup studies for benchmark"""
        with OrionState():
            benchmark.setup_studies()

            assert (
                str(benchmark.studies)
                == "[Study(assessment=AverageResult, task=RosenBrock, algorithms=[random,tpe]), "
                "Study(assessment=AverageResult, task=CarromTable, algorithms=[random,tpe]), "
                "Study(assessment=AverageRank, task=RosenBrock, algorithms=[random,tpe]), "
                "Study(assessment=AverageRank, task=CarromTable, algorithms=[random,tpe])]"
            )

        assert len(benchmark.studies) == 4

    def test_process(self, benchmark, study):
        """Test to process a benchmark"""
        with OrionState():
            study.setup_experiments()
            benchmark.studies = [study]
            benchmark.process()
            name = "benchmark007_AverageResult_RosenBrock_0_0"
            experiment = experiment_builder.build(name)

            assert experiment is not None

    @pytest.mark.usefixtures("version_XYZ")
    def test_status(
        self,
        benchmark,
        study,
        study_experiments_config,
        task_number,
        max_trial,
    ):
        """Test to get the status of a benchmark"""
        with create_study_experiments(**study_experiments_config) as experiments:

            study.experiments_info = experiments

            benchmark.studies = [study]

            assert benchmark.status() == [
                {
                    "Algorithms": "random",
                    "Assessments": "AverageResult",
                    "Tasks": "RosenBrock",
                    "Total Experiments": task_number,
                    "Completed Experiments": task_number,
                    "Submitted Trials": task_number * max_trial,
                },
                {
                    "Algorithms": "tpe",
                    "Assessments": "AverageResult",
                    "Tasks": "RosenBrock",
                    "Total Experiments": task_number,
                    "Completed Experiments": task_number,
                    "Submitted Trials": task_number * max_trial,
                },
            ]

    @pytest.mark.usefixtures("version_XYZ")
    def test_analysis(self, benchmark, study, study_experiments_config):
        """Test to analysis benchmark result"""
        with create_study_experiments(**study_experiments_config) as experiments:

            study.experiments_info = experiments

            benchmark.studies = [study]

            figures = benchmark.analysis()

            assert len(figures) == 1
            assert type(figures[0]) is plotly.graph_objects.Figure

    @pytest.mark.usefixtures("version_XYZ")
    def test_experiments(
        self,
        benchmark,
        study,
        study_experiments_config,
        max_trial,
    ):
        """Test to get experiments list of a benchmark"""
        with create_study_experiments(**study_experiments_config) as experiments:

            study.experiments_info = experiments

            benchmark.studies = [study]

            assert benchmark.experiments() == [
                {
                    "Algorithm": "random",
                    "Experiment Name": "experiment-name-0",
                    "Number Trial": max_trial,
                    "Best Evaluation": 0,
                },
                {
                    "Algorithm": "tpe",
                    "Experiment Name": "experiment-name-1",
                    "Number Trial": max_trial,
                    "Best Evaluation": 0,
                },
                {
                    "Algorithm": "random",
                    "Experiment Name": "experiment-name-2",
                    "Number Trial": max_trial,
                    "Best Evaluation": 0,
                },
                {
                    "Algorithm": "tpe",
                    "Experiment Name": "experiment-name-3",
                    "Number Trial": max_trial,
                    "Best Evaluation": 0,
                },
            ]


class TestStudy:
    """Test Study"""

    def test_creation(self, study):
        """Test study instance creation"""
        assert (
            str(study) == "Study(assessment=AverageResult, "
            "task=RosenBrock, algorithms=[random,tpe])"
        )

    def test_creation_algorithms(self, benchmark):
        """Test study creation with all support algorithms input format"""

        algorithms = [
            {"algorithm": {"gridsearch": {"n_values": 1}}, "deterministic": True},
            {"algorithm": "tpe"},
            {"random": {"seed": 1}},
            "asha",
        ]
        study = Study(benchmark, algorithms, AverageResult(2), RosenBrock(25, dim=3))
        assert study.algorithms[0].name == "gridsearch"
        assert study.algorithms[0].experiment_algorithm == {
            "gridsearch": {"n_values": 1}
        }
        assert study.algorithms[0].is_deterministic

        assert study.algorithms[1].name == "tpe"
        assert study.algorithms[1].experiment_algorithm == "tpe"
        assert not study.algorithms[1].is_deterministic

        assert study.algorithms[2].name == "random"
        assert study.algorithms[2].experiment_algorithm == {"random": {"seed": 1}}
        assert not study.algorithms[2].is_deterministic

        assert study.algorithms[3].name == "asha"
        assert study.algorithms[3].experiment_algorithm == "asha"
        assert not study.algorithms[3].is_deterministic

    def test_setup_experiments(self, study):
        """Test to setup experiments for study"""
        with OrionState():
            study.setup_experiments()

            assert len(study.experiments_info) == 4
            assert isinstance(study.experiments_info[0][1], ExperimentClient)

    def test_execute(self, study):
        """Test to execute a study"""
        with OrionState():
            study.setup_experiments()
            study.execute()
            name = "benchmark007_AverageResult_RosenBrock_0_0"
            experiment = experiment_builder.build(name)

            assert experiment is not None

    @pytest.mark.usefixtures("version_XYZ")
    def test_status(
        self,
        study,
        study_experiments_config,
        task_number,
        max_trial,
    ):
        """Test to get status of a study"""
        with create_study_experiments(**study_experiments_config) as experiments:

            study.experiments_info = experiments

            assert study.status() == [
                {
                    "algorithm": "random",
                    "assessment": "AverageResult",
                    "task": "RosenBrock",
                    "experiments": task_number,
                    "completed": task_number,
                    "trials": task_number * max_trial,
                },
                {
                    "algorithm": "tpe",
                    "assessment": "AverageResult",
                    "task": "RosenBrock",
                    "experiments": task_number,
                    "completed": task_number,
                    "trials": task_number * max_trial,
                },
            ]

    @pytest.mark.usefixtures("version_XYZ")
    def test_analysis(
        self,
        study,
        study_experiments_config,
    ):
        """Test to get the ploty figure of a study"""
        with create_study_experiments(**study_experiments_config) as experiments:

            study.experiments_info = experiments

            plot = study.analysis()

            assert type(plot) is plotly.graph_objects.Figure

    def test_experiments(self, study, study_experiments_config, task_number):
        """Test to get experiments of a study"""
        algo_num = len(study_experiments_config["algorithms"])
        with create_study_experiments(**study_experiments_config) as experiments:

            study.experiments_info = experiments

            experiments = study.experiments()

            assert (
                len(experiments) == study_experiments_config["task_number"] * algo_num
            )
            assert isinstance(experiments[0], Experiment)



# -*- coding: utf-8 -*-
"""Tests :func:`orion.analysis.lpi`"""
import copy

import numpy
import pandas as pd
import pytest

from orion.analysis.base import to_numpy, train_regressor
from orion.analysis.lpi_utils import compute_variances, lpi, make_grid
from orion.core.io.space_builder import SpaceBuilder

data = pd.DataFrame(
    data={
        "id": ["a", "b", "c", "d"],
        "x": [0, 1, 2, 3],
        "y": [1, 2, 0, 3],
        "objective": [0.1, 0.2, 0.3, 0.5],
    }
)


space = SpaceBuilder().build({"x": "uniform(0, 6)", "y": "uniform(0, 3)"})


def test_accept_empty():
    """Tests an empty dataframe is returned if you give an empty dataframe"""
    empty_frame = pd.DataFrame()
    results = lpi(empty_frame, space)

    assert results.columns.tolist() == ["LPI"]
    assert results.index.tolist() == list(space.keys())
    assert results["LPI"].tolist() == [0, 0]

    empty_frame = pd.DataFrame(columns=["x", "y", "objective"])
    results = lpi(empty_frame, space)

    assert results.columns.tolist() == ["LPI"]
    assert results.index.tolist() == list(space.keys())
    assert results["LPI"].tolist() == [0, 0]


def test_parameter_not_modified():
    """Tests the original dataframe is not modified"""
    original = copy.deepcopy(data)
    lpi(data, space)

    pd.testing.assert_frame_equal(data, original)


def test_make_grid():
    """Test grid has correct format"""
    trials = to_numpy(data, space)
    model = train_regressor("RandomForestRegressor", trials)
    best_point = trials[numpy.argmin(trials[:, -1])]
    grid = make_grid(best_point, space, model, 4)

    # Are fixed to anchor value
    numpy.testing.assert_equal(grid[0][:, 1], best_point[1])
    numpy.testing.assert_equal(grid[1][:, 0], best_point[0])

    # Is a grid in search space
    numpy.testing.assert_equal(grid[0][:, 0], [0, 2, 4, 6])
    numpy.testing.assert_equal(grid[1][:, 1], [0, 1, 2, 3])


def test_make_grid_predictor(monkeypatch):
    """Test grid contains corresponding predictions from the model"""
    trials = to_numpy(data, space)
    model = train_regressor("RandomForestRegressor", trials)
    best_point = trials[numpy.argmin(trials[:, -1])]

    # Make sure model is not predicting exactly the original objective
    with numpy.testing.assert_raises(AssertionError):
        numpy.testing.assert_equal(
            best_point[-1], model.predict(best_point[:-1].reshape(1, -1))
        )

    grid = make_grid(best_point, space, model, 4)

    # Verify that grid predictions are those of the model
    numpy.testing.assert_equal(grid[0][:, -1], model.predict(grid[0][:, :-1]))
    numpy.testing.assert_equal(grid[1][:, -1], model.predict(grid[1][:, :-1]))

    # Verify model predictions differ on different points
    with numpy.testing.assert_raises(AssertionError):
        numpy.testing.assert_equal(grid[0][:, -1], grid[1][:, -1])


def test_compute_variance():
    """Test variance computation over the grid"""
    grid = numpy.arange(3 * 5 * 4).reshape(3, 5, 4)

    grid[0, :, -1] = 10
    grid[1, :, -1] = [0, 1, 2, 3, 4]
    grid[2, :, -1] = [0, 10, 20, 30, 40]

    variances = compute_variances(grid)
    assert variances.shape == (3,)
    assert variances[0] == 0
    assert variances[1] == numpy.var([0, 1, 2, 3, 4])
    assert variances[2] == numpy.var([0, 10, 20, 30, 40])


def test_lpi_results():
    """Verify LPI results in DataFrame"""
    results = lpi(data, space, random_state=1)
    assert results.columns.tolist() == ["LPI", "STD"]
    assert results.index.tolist() == list(space.keys())
    # The data is made such that x correlates more strongly with objective than y
    assert results["LPI"].loc["x"] > results["LPI"].loc["y"]


def test_lpi_with_categorical_data():
    """Verify LPI can be computed on categorical dimensions"""
    data = pd.DataFrame(
        data={
            "id": ["a", "b", "c", "d"],
            "x": [0, 1, 2, 3],
            "y": ["b", "c", "a", "d"],
            "objective": [0.1, 0.2, 0.3, 0.5],
        }
    )

    space = SpaceBuilder().build(
        {"x": "uniform(0, 6)", "y": 'choices(["a", "b", "c", "d"])'}
    )

    results = lpi(data, space, random_state=1)
    assert results.columns.tolist() == ["LPI", "STD"]
    assert results.index.tolist() == ["x", "y"]
    # The data is made such that x correlates more strongly with objective than y
    assert results["LPI"].loc["x"] > results["LPI"].loc["y"]


def test_lpi_with_multidim_data():
    """Verify LPI can be computed on categorical dimensions"""
    data = pd.DataFrame(
        data={
            "id": ["a", "b", "c", "d"],
            "x": [[0, 2, 4], [1, 1, 3], [2, 2, 2], [3, 0, 3]],
            "y": [["b", "b"], ["c", "b"], ["a", "a"], ["d", "c"]],
            "objective": [0.1, 0.2, 0.3, 0.5],
        }
    )

    space = SpaceBuilder().build(
        {"x": "uniform(0, 6, shape=3)", "y": 'choices(["a", "b", "c", "d"], shape=2)'}
    )

    results = lpi(data, space, random_state=1)
    assert results.columns.tolist() == ["LPI", "STD"]
    assert results.index.tolist() == ["x[0]", "x[1]", "x[2]", "y[0]", "y[1]"]
    # The data is made such some x correlates more strongly with objective than other x and most y
    assert results["LPI"].loc["x[0]"] > results["LPI"].loc["x[1]"]
    assert results["LPI"].loc["x[1]"] > results["LPI"].loc["x[2]"]
    assert results["LPI"].loc["x[0]"] > results["LPI"].loc["y[0]"]
    assert results["LPI"].loc["x[0]"] > results["LPI"].loc["y[1]"]


def test_lpi_n_points(monkeypatch):
    """Verify given number of points is used"""
    N_POINTS = numpy.random.randint(2, 50)

    def mock_make_grid(*args, **kwargs):
        grid = make_grid(*args, **kwargs)
        assert grid.shape == (len(space), N_POINTS, len(space) + 1)
        return grid

    monkeypatch.setattr("orion.analysis.lpi_utils.make_grid", mock_make_grid)
    lpi(data, space, random_state=1, n_points=N_POINTS)


def test_lpi_n_runs(monkeypatch):
    """Verify number of runs"""
    N_RUNS = 5

    seeds = set()
    n_runs = 0

    def mock_train_regressor(*args, **kwargs):
        nonlocal n_runs
        n_runs += 1
        seeds.add(kwargs["random_state"])
        return train_regressor(*args, **kwargs)

    monkeypatch.setattr(
        "orion.analysis.lpi_utils.train_regressor", mock_train_regressor
    )
    lpi(data, space, random_state=1, n_runs=N_RUNS)

    assert n_runs == N_RUNS
    assert len(seeds) > 0



"""
============
How to plot.
============

There exist multiple ways to plot using Oríon. We will start with the most convenient one, the
experiment object plot accessor.


Accessor ``experiment.plot``
----------------------------

.. note::

    It only supports single experiment plots. It does not support ``regrets`` and ``averages``.

It is possible to render most plots directly with 
the :py:class:`ExperimentClient <orion.client.experiment.ExperimentClient>` object.
The accessor ``ExperimentClient.plot`` can be used to plot the results of the experiment.

"""
from orion.client import get_experiment

# Specify the database where the experiments are stored. We use a local PickleDB here.
storage = dict(type="legacy", database=dict(type="pickleddb", host="../../db.pkl"))

# Load the data for the specified experiment
experiment = get_experiment("2-dim-exp", storage=storage)
fig = experiment.plot.regret()
fig.show()

#%%
# Module ``plotting``
# -------------------
#
# The plotting module can used as well to plot experiments.

import orion.plotting.base as plot

fig = plot.regret(experiment)
fig.show()

#%%
# The advantage of the plotting module is that it can render plots using
# multiple examples as shown in the example below.
#

fig = plot.regret(experiment)
fig.show()

# sphinx_gallery_thumbnail_path = '_static/paint.png'

#%%
# Command ``orion plot``
# ----------------------
#
# .. note::
#
#     The plotting command line utility does not support arguments for now. Contributions are
#     welcome! :)
#
# .. note::
#
#     Only supports single experiment plots. It does not support ``regrets`` and ``averages``.
#
# This section refers to the use of the ``orion plot ...`` subcommand.
#
# Usage
# ~~~~~
#
# The arguments expected by the "plot" subcommand are as follows
#
# .. code-block :: console
#
#    usage: orion plot [-h] [-n stringID] [-u USER] [-v VERSION] [-c path-to-config]
#                      [-t {png,jpg,jpeg,webp,svg,pdf,html,json}]
#                      [-o OUTPUT] [--scale SCALE]
#                      {lpi,parallel_coordinates,partial_dependencies,regret}
#
#    Produce plots for Oríon experiments
#
#    positional arguments:
#    {lpi,parallel_coordinates,partial_dependencies,regret}
#                            kind of plot to generate.
#
#    optional arguments:
#    -h, --help            show this help message and exit
#    -t {png,jpg,jpeg,webp,svg,pdf,html,json}, --type {png,jpg,jpeg,webp,svg,pdf,html,json}
#                            type of plot to return. (default: png)
#    -o OUTPUT, --output OUTPUT
#                            path where plot is saved. Will override the `type` argument.
#                            (default is {exp.name}-v{exp.version}_{kind}.{type})
#    --scale SCALE         more pixels, but same proportions of the plot.
#                            Scale acts as multiplier on height and width of resulting image.
#                            Overrides value of 'scale' in plotly.io.write_image.
#
#    Oríon arguments:
#    These arguments determine orion's behaviour
#
#    -n stringID, --name stringID
#                            experiment's unique name;
#                            (default: None - specified either here or in a config)
#    -u USER, --user USER  user associated to experiment's unique name;
#                            (default: $USER - can be overriden either here or in a config)
#    -v VERSION, --version VERSION
#                            specific version of experiment to fetch; (default: None - latest experiment.)
#    -c path-to-config, --config path-to-config
#                            user provided orion configuration file
#
# More about the arguments
# ~~~~~~~~~~~~~~~~~~~~~~~~
#
# The ``kind`` positional argument is the only mandatory argument.
# Assuming that the user identified the experiment in the usual
# way (e.g. using ``--name`` or a config file), the default behavior
# is to generate the correct ``kind`` of plot, and to save it
# as a "png" file in the current directory and with a filename
# automatically formatted as "{experiment.name}_{kind}.png".
#
# kind
# ^^^^
#
# The plotting command requires a ``kind`` argument to determine which
# of the four kinds of plots to generate.
# The choice is between 'lpi', 'partial_dependencies', 'parallel_coordinates'
# or 'regret'.
#
#
# type
# ^^^^
#
# The ``type`` is basically the filename extension. This governs more than just the name of the file
# because it determines the actual format of the output. The default is to give the user a 'png' file.
#
# Behind the scenes, *plotly* generates an initial 'json' file, and renders it as an image
# to be saved in the desired format. With ``type`` being 'json', that original file
# is saved without rendering it to an image.
#
# The accepted values of ``type`` are 'png', 'jpg', 'jpeg', 'webp', 'svg', 'pdf', 'html' and 'json'.
#
# output
# ^^^^^^
#
# The default value for the output filename is an automatically-generated
# string formatted as "{experiment.name}_{kind}.{type}".
# This also implies that the plot will be saved in the current directory.
#
# When ``output`` is specified by the user with a file extension, the ``type`` argument
# will be ignored because the value of ``type`` will be instead inferred by
# the file extension in the ``output``.
# For example, with ``--output=../myplot.jpg``, the results will be saved
# in the parent directory and the ``type`` will be 'jpg'.
#
# scale
# ^^^^^
#
# With certain types of plots, it can be desirable to increase the
# resolution of the output image in terms of pixel counts (equivalent to dpi).
# This applies particularly to 'jpg' and 'png', but it does not affect 'json' or 'html'.
#
# The reference value of ``scale`` is 1.0.
# With ``--scale 2.0``, the height and width are going to be doubled.
#
# Web API
# -------
#
# .. note::
#
#     Only supports single experiment plots. It does not support ``regrets`` and ``averages``.
#
# The :ref:`web-api` supports queries for single experiment plots. See documentation for
# all queries. The JSON output is generated automatically according to the
# `Plotly.js schema reference <https://plotly.com/python/reference/index/>`_.



"""Example usage and tests for :mod:`orion.algo.gridsearch`."""
import copy
import logging

import numpy.testing
import pytest

from orion.algo.gridsearch import (
    GridSearch,
    categorical_grid,
    discrete_grid,
    grid,
    real_grid,
)
from orion.algo.space import Categorical, Integer, Real, Space
from orion.testing.algo import BaseAlgoTests, phase


def test_categorical_grid():
    """Test that categorical grid returns all choices"""
    dim = Categorical("yolo", "abcde")
    assert categorical_grid(dim, 5) == dim.categories


def test_too_large_categorical_grid(caplog):
    """Test that large categorical grid returns all choices"""
    dim = Categorical("yolo", "abcde")
    with caplog.at_level(logging.WARNING, logger="orion.algo.gridsearch"):
        assert categorical_grid(dim, 3) == dim.categories
    assert "Categorical dimension yolo does not have 3 choices" in caplog.text


def test_too_small_categorical_grid(caplog):
    """Test that small categorical grid returns all choices"""
    dim = Categorical("yolo", "abcde")
    with caplog.at_level(logging.WARNING, logger="orion.algo.gridsearch"):
        assert categorical_grid(dim, 7) == dim.categories
    assert "Categorical dimension yolo does not have 7 choices" in caplog.text


def test_discrete_grid():
    """Test discrete grid"""
    dim = Integer("yolo", "uniform", -5, 10)
    assert discrete_grid(dim, 6) == [-5, -3, -1, 1, 3, 5]
    assert discrete_grid(dim, 7) == [-5, -3, -2, 0, 2, 3, 5]


def test_too_small_distrete_grid():
    """Test that small discrete grid does not lead to duplicates"""
    dim = Integer("yolo", "uniform", -2, 4)
    assert discrete_grid(dim, 3) == [-2, 0, 2]
    assert discrete_grid(dim, 5) == [-2, -1, 0, 1, 2]
    assert discrete_grid(dim, 50) == [-2, -1, 0, 1, 2]


def test_log_discrete_grid():
    """Test log discrete grid"""
    dim = Integer("yolo", "reciprocal", 1, 1000)
    assert discrete_grid(dim, 4) == [1, 10, 100, 1000]
    assert discrete_grid(dim, 6) == [1, 4, 16, 63, 251, 1000]


def test_unsupported_discrete_prior():
    """Test unsupported discrete prior message"""
    dim = Integer("yolo", "norm", 0, 1)
    with pytest.raises(TypeError) as exc:
        discrete_grid(dim, 6)

    assert exc.match("Grid Search only supports `loguniform`, `uniform`")


def test_real_grid():
    """Test linear real grid"""
    dim = Real("yolo", "uniform", -5, 10)
    assert real_grid(dim, 6) == [-5, -3, -1, 1, 3, 5]
    assert real_grid(dim, 7) == [-5 + 10 / 6.0 * i for i in range(7)]


def test_log_real_grid():
    """Test logarithmic real grid"""
    dim = Real("yolo", "reciprocal", 0.0001, 1)
    assert numpy.allclose(real_grid(dim, 5), [0.0001, 0.001, 0.01, 0.1, 1])
    assert real_grid(dim, 6) == [
        numpy.exp(numpy.log(0.0001) + (numpy.log(1) - numpy.log(0.0001)) / 5 * i)
        for i in range(6)
    ]


def test_unsupported_real_prior():
    """Test unsupported real prior message"""
    dim = Real("yolo", "norm", 0.0001, 1)
    with pytest.raises(TypeError) as exc:
        real_grid(dim, 5)

    assert exc.match("Grid Search only supports `loguniform`, `uniform`")


def test_unsupported_type():
    """Test unsupported real prior message"""

    class NewDimType:
        @property
        def type(self):
            return "newdimtype"

    dim = NewDimType()
    with pytest.raises(TypeError) as exc:
        grid(dim, 5)

    assert exc.match(
        f"Grid Search only supports `real`, `integer`, `categorical` and `fidelity`: `{dim.type}`"
    )


def test_build_grid():
    """Test that grid search builds the proper grid"""
    dim1 = Real("dim1", "uniform", 0, 1)
    dim2 = Integer("dim2", "uniform", 0, 10)
    dim3 = Categorical("dim3", "abcde")
    space = Space()
    space.register(dim1)
    space.register(dim2)
    space.register(dim3)

    grid = GridSearch.build_grid(space, {"dim1": 3, "dim2": 4, "dim3": 1})
    assert len(grid) == 3 * 4 * 5


def test_build_grid_limit_size(caplog):
    """Test that grid search reduces the n_values when grid is too large"""
    dim1 = Real("dim1", "uniform", 0, 1)
    dim2 = Integer("dim2", "uniform", 0, 10)
    dim3 = Categorical("dim3", "abcde")
    space = Space()
    space.register(dim1)
    space.register(dim2)
    space.register(dim3)

    with caplog.at_level(logging.WARNING, logger="orion.algo.gridsearch"):
        grid = GridSearch.build_grid(space, {k: 5 for k in space}, 100)
    assert len(grid) == 4 * 4 * 5

    assert "`n_values` reduced by 1 to limit number of trials below 100" in caplog.text


def test_build_grid_cannot_limit_size(caplog):
    """Test that when choices are too large GridSearch raises ValueError"""
    dim1 = Real("dim1", "uniform", 0, 1)
    dim2 = Integer("dim2", "uniform", 0, 10)
    dim3 = Categorical("dim3", "abcde")
    dim4 = Categorical("dim4", "abcde")
    space = Space()
    space.register(dim1)
    space.register(dim2)
    space.register(dim3)
    space.register(dim4)

    with pytest.raises(ValueError) as exc:
        GridSearch.build_grid(space, {k: 5 for k in space}, 10)

    assert exc.match(
        "Cannot build a grid smaller than 10. "
        "Try reducing the number of choices in categorical dimensions."
    )


class TestGridSearch(BaseAlgoTests):
    algo_name = "gridsearch"
    config = {"n_values": 10}

    @phase
    def test_suggest_lots(self, mocker, num, attr):
        """Test that gridsearch returns the whole grid when requesting more points"""
        algo = self.create_algo()
        spy = self.spy_phase(mocker, num, algo, attr)
        points = algo.suggest(10000)
        assert len(points) == len(algo.algorithm.grid)

    @pytest.mark.skip(reason="Deterministic algorithm")
    def test_seed_rng(self, mocker, num, attr):
        pass


TestGridSearch.set_phases([("grid", 0, "suggest")])



"""
Script to turn the database ``examples/plotting/database.pkl`` into a clean
version ``examples/base_db.pkl`` for the examples.
"""
import pprint
import shutil

from orion.client import get_experiment
from orion.core.io.orion_cmdline_parser import OrionCmdlineParser
from orion.storage.base import get_storage, setup_storage


shutil.copy("./examples/plotting/database.pkl", "./examples/base_db.pkl")

setup_storage(
    dict(
        type="legacy",
        database=dict(type="pickleddb", host="./examples/base_db.pkl"),
    )
)

filter_exps = {
    ("lateral-view-pa4", 1): "2-dim-exp",
    ("lateral-view-dualnet2", 1): "2-dim-shape-exp",
    ("lateral-view-multitask2", 1): "4-dim-cat-shape-exp",
    ("lateral-view-multitask3", 1): "3-dim-cat-shape-exp",
}

storage = get_storage()


def update_dropout(experiment_config):
    metadata = experiment_config["metadata"]
    user_script = metadata.get("user_script", "")
    user_args = metadata.get("user_args", [])
    try:
        index = user_args.index("--dropout")
    except ValueError:
        print(
            f"No dropout for {experiment_config['metadata']}-v{experiment_config['version']}"
        )
        return

    user_args[index + 1] = (
        user_args[index + 1]
        .replace("5,", "0.5,")
        .replace(", discrete=True", ", precision=1")
    )
    cmdline_parser = OrionCmdlineParser(allow_non_existing_files=True)
    cmdline_parser.parse([user_script] + user_args)
    metadata["parser"] = cmdline_parser.get_state_dict()
    experiment_config["space"] = metadata["priors"] = dict(cmdline_parser.priors)

    # Update config in db
    storage.update_experiment(uid=experiment_config["_id"], **experiment_config)

    # Update all trials in db (arf)
    n_trials_before = len(storage.fetch_trials(uid=experiment_config["_id"]))
    for trial in storage.fetch_trials(uid=experiment_config["_id"]):
        previous_id = trial.id
        for param in trial._params:
            if param.name == "/dropout":
                param.value /= 10
                assert 0 <= param.value <= 0.5, param.value

        storage.delete_trials(uid=experiment_config["_id"], where=dict(_id=previous_id))
        storage.register_trial(trial)

    trials = storage.fetch_trials(uid=experiment_config["_id"])
    assert len(trials) == n_trials_before, len(trials)
    for trial in trials:
        assert 0 <= trial.params["/dropout"] <= 0.5, trial


for experiment_config in storage.fetch_experiments({}):
    key = (experiment_config["name"], experiment_config["version"])

    if key in filter_exps:
        print(
            f"Saving {experiment_config['name']}-v{experiment_config['version']} as {filter_exps[key]}"
        )
        update_dropout(experiment_config)

        storage.update_experiment(
            uid=experiment_config["_id"], name=filter_exps[key], version=1
        )
    else:
        print(f"Deleting {experiment_config['name']}-v{experiment_config['version']}")
        storage.delete_experiment(uid=experiment_config["_id"])
        storage._db.remove("trials", query={"experiment": experiment_config["_id"]})



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Collection of tests for :mod:`orion.core.worker.experiment`."""

import copy
import datetime
import inspect
import json
import logging
import pickle
import tempfile

import pandas
import pytest

import orion.core
import orion.core.utils.backward as backward
import orion.core.worker.experiment
from orion.core.io.space_builder import SpaceBuilder
from orion.core.utils.exceptions import UnsupportedOperation
from orion.core.worker.experiment import Experiment
from orion.core.worker.primary_algo import PrimaryAlgo
from orion.core.worker.trial import Trial
from orion.storage.base import get_storage
from orion.testing import OrionState


@pytest.fixture()
def new_config(random_dt):
    """Create a configuration that will not hit the database."""
    new_config = dict(
        name="supernaekei",
        # refers is missing on purpose
        metadata={
            "user": "tsirif",
            "orion_version": 0.1,
            "user_script": "tests/functional/demo/black_box.py",
            "user_config": "abs_path/hereitis.yaml",
            "user_args": ["--mini-batch~uniform(32, 256, discrete=True)"],
            "VCS": {
                "type": "git",
                "is_dirty": False,
                "HEAD_sha": "fsa7df7a8sdf7a8s7",
                "active_branch": None,
                "diff_sha": "diff",
            },
        },
        version=1,
        pool_size=10,
        max_trials=1000,
        max_broken=5,
        working_dir=None,
        algorithms={"dumbalgo": {}},
        producer={"strategy": "NoParallelStrategy"},
        # attrs starting with '_' also
        # _id='fasdfasfa',
        # and in general anything which is not in Experiment's slots
        something_to_be_ignored="asdfa",
    )

    backward.populate_space(new_config)

    return new_config


@pytest.fixture
def parent_version_config():
    """Return a configuration for an experiment."""
    config = dict(
        _id="parent_config",
        name="old_experiment",
        version=1,
        algorithms="random",
        metadata={
            "user": "corneauf",
            "datetime": datetime.datetime.utcnow(),
            "user_args": ["--x~normal(0,1)"],
        },
    )

    backward.populate_space(config)

    return config


@pytest.fixture
def child_version_config(parent_version_config):
    """Return a configuration for an experiment."""
    config = copy.deepcopy(parent_version_config)
    config["_id"] = "child_config"
    config["version"] = 2
    config["refers"] = {"parent_id": "parent_config"}
    config["metadata"]["datetime"] = datetime.datetime.utcnow()
    config["metadata"]["user_args"].append("--y~+normal(0,1)")
    backward.populate_space(config)
    return config


def _generate(obj, *args, value):
    if obj is None:
        return None

    obj = copy.deepcopy(obj)
    data = obj

    for arg in args[:-1]:
        data = data[arg]

    data[args[-1]] = value
    return obj


base_trial = {
    "experiment": 0,
    "status": "new",  # new, reserved, suspended, completed, broken
    "worker": None,
    "submit_time": "2017-11-23T02:00:00",
    "start_time": None,
    "end_time": None,
    "heartbeat": None,
    "results": [
        {"name": "loss", "type": "objective", "value": 2}  # objective, constraint
    ],
    "params": [
        {"name": "/encoding_layer", "type": "categorical", "value": "rnn"},
        {
            "name": "/decoding_layer",
            "type": "categorical",
            "value": "lstm_with_attention",
        },
    ],
}


def generate_trials(status):
    """Generate Trials with different configurations"""
    new_trials = [_generate(base_trial, "status", value=s) for s in status]

    for i, trial in enumerate(new_trials):
        if trial["status"] != "new":
            trial["start_time"] = datetime.datetime.utcnow() + datetime.timedelta(
                seconds=i
            )

    for i, trial in enumerate(new_trials):
        if trial["status"] == "completed":
            trial["end_time"] = datetime.datetime.utcnow() + datetime.timedelta(
                seconds=i
            )

    # make each trial unique
    for i, trial in enumerate(new_trials):
        trial["results"][0]["value"] = i
        trial["params"].append({"name": "/index", "type": "categorical", "value": i})

    return new_trials


def assert_protocol(exp, storage):
    """Transitional method to move away from mongodb"""
    assert exp._storage._db is storage._db


def count_experiment(exp):
    """Transitional method to move away from mongodb"""
    return exp._storage._db.count("experiments")


@pytest.fixture()
def space():
    """Build a space object"""
    return SpaceBuilder().build({"/index": "uniform(0, 10)"})


@pytest.fixture()
def algorithm(space):
    """Build a dumb algo object"""
    return PrimaryAlgo(space, "dumbalgo")


class TestReserveTrial(object):
    """Calls to interface `Experiment.reserve_trial`."""

    @pytest.mark.usefixtures("setup_pickleddb_database")
    def test_reserve_none(self):
        """Find nothing, return None."""
        with OrionState(experiments=[], trials=[]):
            exp = Experiment("supernaekei", mode="x")
            trial = exp.reserve_trial()
            assert trial is None

    def test_reserve_success(self, random_dt):
        """Successfully find new trials in db and reserve the first one"""
        storage_config = {"type": "legacy", "database": {"type": "EphemeralDB"}}
        with OrionState(
            trials=generate_trials(["new", "reserved"]), storage=storage_config
        ) as cfg:
            exp = Experiment("supernaekei", mode="x")
            exp._id = cfg.trials[0]["experiment"]

            trial = exp.reserve_trial()

            # Trials are sorted according to hash and 'new' gets position second
            cfg.trials[1]["status"] = "reserved"
            cfg.trials[1]["start_time"] = random_dt
            cfg.trials[1]["heartbeat"] = random_dt

            assert trial.to_dict() == cfg.trials[1]

    def test_reserve_when_exhausted(self):
        """Return None once all the trials have been allocated"""
        stati = ["new", "reserved", "interrupted", "completed", "broken"]
        with OrionState(trials=generate_trials(stati)) as cfg:
            exp = Experiment("supernaekei", mode="x")
            exp._id = cfg.trials[0]["experiment"]
            assert exp.reserve_trial() is not None
            assert exp.reserve_trial() is not None
            assert exp.reserve_trial() is None

    def test_fix_lost_trials(self):
        """Test that a running trial with an old heartbeat is set to interrupted."""
        trial = copy.deepcopy(base_trial)
        trial["status"] = "reserved"
        trial["heartbeat"] = datetime.datetime.utcnow() - datetime.timedelta(
            seconds=60 * 10
        )
        with OrionState(trials=[trial]) as cfg:
            exp = Experiment("supernaekei", mode="x")
            exp._id = cfg.trials[0]["experiment"]

            assert len(exp.fetch_trials_by_status("reserved")) == 1
            exp.fix_lost_trials()
            assert len(exp.fetch_trials_by_status("reserved")) == 0

    def test_fix_only_lost_trials(self):
        """Test that an old trial is set to interrupted but not a recent one."""
        lost_trial, running_trial = generate_trials(["reserved"] * 2)
        lost_trial["heartbeat"] = datetime.datetime.utcnow() - datetime.timedelta(
            seconds=60 * 10
        )
        running_trial["heartbeat"] = datetime.datetime.utcnow()

        with OrionState(trials=[lost_trial, running_trial]) as cfg:
            exp = Experiment("supernaekei", mode="x")
            exp._id = cfg.trials[0]["experiment"]

            assert len(exp.fetch_trials_by_status("reserved")) == 2

            exp.fix_lost_trials()

            reserved_trials = exp.fetch_trials_by_status("reserved")
            assert len(reserved_trials) == 1
            assert reserved_trials[0].to_dict()["params"] == running_trial["params"]

            failedover_trials = exp.fetch_trials_by_status("interrupted")
            assert len(failedover_trials) == 1
            assert failedover_trials[0].to_dict()["params"] == lost_trial["params"]

    def test_fix_lost_trials_race_condition(self, monkeypatch, caplog):
        """Test that a lost trial fixed by a concurrent process does not cause error."""
        trial = copy.deepcopy(base_trial)
        trial["status"] = "interrupted"
        trial["heartbeat"] = datetime.datetime.utcnow() - datetime.timedelta(
            seconds=60 * 10
        )
        with OrionState(trials=[trial]) as cfg:
            exp = Experiment("supernaekei", mode="x")
            exp._id = cfg.trials[0]["experiment"]

            assert len(exp.fetch_trials_by_status("interrupted")) == 1

            assert len(exp._storage.fetch_lost_trials(exp)) == 0

            def fetch_lost_trials(self, query):
                trial_object = Trial(**trial)
                trial_object.status = "reserved"
                return [trial_object]

            # Force the fetch of a trial marked as reserved (and lost) while actually interrupted
            # (as if already failed-over by another process).
            with monkeypatch.context() as m:
                m.setattr(
                    exp._storage.__class__, "fetch_lost_trials", fetch_lost_trials
                )

                assert len(exp._storage.fetch_lost_trials(exp)) == 1

                with caplog.at_level(logging.DEBUG):
                    exp.fix_lost_trials()

            assert caplog.records[-1].levelname == "DEBUG"
            assert caplog.records[-1].msg == "failed"
            assert len(exp.fetch_trials_by_status("interrupted")) == 1
            assert len(exp.fetch_trials_by_status("reserved")) == 0

    def test_fix_lost_trials_configurable_hb(self):
        """Test that heartbeat is correctly being configured."""
        trial = copy.deepcopy(base_trial)
        trial["status"] = "reserved"
        trial["heartbeat"] = datetime.datetime.utcnow() - datetime.timedelta(
            seconds=60 * 2
        )
        with OrionState(trials=[trial]) as cfg:
            exp = Experiment("supernaekei", mode="x")
            exp._id = cfg.trials[0]["experiment"]

            assert len(exp.fetch_trials_by_status("reserved")) == 1

            orion.core.config.worker.heartbeat = 60 * 2

            exp.fix_lost_trials()

            assert len(exp.fetch_trials_by_status("reserved")) == 1

            orion.core.config.worker.heartbeat = 60 * 2 / 10.0

            exp.fix_lost_trials()

            assert len(exp.fetch_trials_by_status("reserved")) == 0


def test_update_completed_trial(random_dt):
    """Successfully push a completed trial into database."""
    with OrionState(trials=generate_trials(["new"])) as cfg:
        exp = Experiment("supernaekei", mode="x")
        exp._id = cfg.trials[0]["experiment"]

        trial = exp.reserve_trial()

        results_file = tempfile.NamedTemporaryFile(
            mode="w", prefix="results_", suffix=".log", dir=".", delete=True
        )

        # Generate fake result
        with open(results_file.name, "w") as file:
            json.dump([{"name": "loss", "type": "objective", "value": 2}], file)
        # --

        exp.update_completed_trial(trial, results_file=results_file)

        yo = get_storage().fetch_trials(exp)[0].to_dict()

        assert len(yo["results"]) == len(trial.results)
        assert yo["results"][0] == trial.results[0].to_dict()
        assert yo["status"] == "completed"
        assert yo["end_time"] == random_dt

        results_file.close()


@pytest.mark.usefixtures("with_user_tsirif")
def test_register_trials(random_dt):
    """Register a list of newly proposed trials/parameters."""
    with OrionState():
        exp = Experiment("supernaekei", mode="x")
        exp._id = 0

        trials = [
            Trial(params=[{"name": "a", "type": "integer", "value": 5}]),
            Trial(params=[{"name": "b", "type": "integer", "value": 6}]),
        ]
        for trial in trials:
            exp.register_trial(trial)

        yo = list(map(lambda trial: trial.to_dict(), get_storage().fetch_trials(exp)))
        assert len(yo) == len(trials)
        assert yo[0]["params"] == list(map(lambda x: x.to_dict(), trials[0]._params))
        assert yo[1]["params"] == list(map(lambda x: x.to_dict(), trials[1]._params))
        assert yo[0]["status"] == "new"
        assert yo[1]["status"] == "new"
        assert yo[0]["submit_time"] == random_dt
        assert yo[1]["submit_time"] == random_dt


class TestToPandas:
    """Test suite for ``Experiment.to_pandas``"""

    def test_empty(self, space):
        """Test panda frame creation when there is no trials"""
        with OrionState():
            exp = Experiment("supernaekei", mode="x")
            exp.space = space
            assert exp.to_pandas().shape == (0, 8)
            assert list(exp.to_pandas().columns) == [
                "id",
                "experiment_id",
                "status",
                "suggested",
                "reserved",
                "completed",
                "objective",
                "/index",
            ]

    def test_data(self, space):
        """Verify the data in the panda frame is coherent with database"""
        with OrionState(
            trials=generate_trials(["new", "reserved", "completed"])
        ) as cfg:
            exp = Experiment("supernaekei", mode="x")
            exp._id = cfg.trials[0]["experiment"]
            exp.space = space
            df = exp.to_pandas()
            assert df.shape == (3, 8)
            assert list(df["id"]) == [trial["_id"] for trial in cfg.trials]
            assert all(df["experiment_id"] == exp._id)
            assert list(df["status"]) == ["completed", "reserved", "new"]
            assert list(df["suggested"]) == [
                trial["submit_time"] for trial in cfg.trials
            ]
            assert df["reserved"][0] == cfg.trials[0]["start_time"]
            assert df["reserved"][1] == cfg.trials[1]["start_time"]
            assert df["reserved"][2] is pandas.NaT
            assert df["completed"][0] == cfg.trials[0]["end_time"]
            assert df["completed"][1] is pandas.NaT
            assert df["completed"][2] is pandas.NaT
            assert list(df["objective"]) == [2, 1, 0]
            assert list(df["/index"]) == [2, 1, 0]


def test_fetch_all_trials():
    """Fetch a list of all trials"""
    with OrionState(trials=generate_trials(["new", "reserved", "completed"])) as cfg:
        exp = Experiment("supernaekei", mode="x")
        exp._id = cfg.trials[0]["experiment"]

        trials = list(map(lambda trial: trial.to_dict(), exp.fetch_trials({})))
        assert trials == cfg.trials


def test_fetch_non_completed_trials():
    """Fetch a list of the trials that are not completed

    trials.status in ['new', 'interrupted', 'suspended', 'broken']
    """
    non_completed_stati = ["new", "interrupted", "suspended", "reserved"]
    stati = non_completed_stati + ["completed"]
    with OrionState(trials=generate_trials(stati)) as cfg:
        exp = Experiment("supernaekei", mode="x")
        exp._id = cfg.trials[0]["experiment"]

        trials = exp.fetch_noncompleted_trials()
        assert len(trials) == 4
        assert set(trial.status for trial in trials) == set(non_completed_stati)


def test_is_done_property_with_pending(algorithm):
    """Check experiment stopping conditions when there is pending trials."""
    completed = ["completed"] * 10
    reserved = ["reserved"] * 5
    with OrionState(trials=generate_trials(completed + reserved)) as cfg:
        exp = Experiment("supernaekei", mode="x")
        exp._id = cfg.trials[0]["experiment"]

        exp.algorithms = algorithm
        exp.max_trials = 10

        assert exp.is_done

        exp.max_trials = 15

        # There is only 10 completed trials
        assert not exp.is_done

        exp.algorithms.algorithm.done = True

        # Algorithm is done but 5 trials are pending
        assert not exp.is_done


def test_is_done_property_no_pending(algorithm):
    """Check experiment stopping conditions when there is no pending trials."""
    completed = ["completed"] * 10
    broken = ["broken"] * 5
    with OrionState(trials=generate_trials(completed + broken)) as cfg:
        exp = Experiment("supernaekei", mode="x")
        exp._id = cfg.trials[0]["experiment"]

        exp.algorithms = algorithm

        exp.max_trials = 15

        # There is only 10 completed trials and algo not done.
        assert not exp.is_done

        exp.algorithms.algorithm.done = True

        # Algorithm is done and no pending trials
        assert exp.is_done


def test_broken_property():
    """Check experiment stopping conditions for maximum number of broken."""
    MAX_BROKEN = 5

    stati = (["reserved"] * 10) + (["broken"] * (MAX_BROKEN - 1))
    with OrionState(trials=generate_trials(stati)) as cfg:
        exp = Experiment("supernaekei", mode="x")
        exp._id = cfg.trials[0]["experiment"]

        exp.max_broken = MAX_BROKEN

        assert not exp.is_broken

    stati = (["reserved"] * 10) + (["broken"] * (MAX_BROKEN))
    with OrionState(trials=generate_trials(stati)) as cfg:
        exp = Experiment("supernaekei", mode="x")
        exp._id = cfg.trials[0]["experiment"]

        exp.max_broken = MAX_BROKEN

        assert exp.is_broken


def test_configurable_broken_property():
    """Check if max_broken changes after configuration."""
    MAX_BROKEN = 5

    stati = (["reserved"] * 10) + (["broken"] * (MAX_BROKEN))
    with OrionState(trials=generate_trials(stati)) as cfg:
        exp = Experiment("supernaekei", mode="x")
        exp._id = cfg.trials[0]["experiment"]

        exp.max_broken = MAX_BROKEN

        assert exp.is_broken

        exp.max_broken += 1

        assert not exp.is_broken


def test_experiment_stats():
    """Check that property stats is returning a proper summary of experiment's results."""
    NUM_COMPLETED = 3
    stati = (["completed"] * NUM_COMPLETED) + (["reserved"] * 2)
    with OrionState(trials=generate_trials(stati)) as cfg:
        exp = Experiment("supernaekei", mode="x")
        exp._id = cfg.trials[0]["experiment"]
        exp.metadata = {"datetime": datetime.datetime.utcnow()}
        stats = exp.stats
        assert stats["trials_completed"] == NUM_COMPLETED
        assert stats["best_trials_id"] == cfg.trials[3]["_id"]
        assert stats["best_evaluation"] == 0
        assert stats["start_time"] == exp.metadata["datetime"]
        assert stats["finish_time"] == cfg.trials[0]["end_time"]
        assert stats["duration"] == stats["finish_time"] - stats["start_time"]
        assert len(stats) == 6


def test_experiment_pickleable():
    """Test experiment instance is pickleable"""

    with OrionState(trials=generate_trials(["new"])) as cfg:
        exp = Experiment("supernaekei", mode="x")
        exp._id = cfg.trials[0]["experiment"]

        exp_trials = exp.fetch_trials()

        assert len(exp_trials) > 0

        exp_bytes = pickle.dumps(exp)

        new_exp = pickle.loads(exp_bytes)

        assert [trial.to_dict() for trial in exp_trials] == [
            trial.to_dict() for trial in new_exp.fetch_trials()
        ]


read_only_methods = [
    "algorithms",
    "configuration",
    "fetch_noncompleted_trials",
    "fetch_trials",
    "fetch_trials_by_status",
    "get_trial",
    "id",
    "is_broken",
    "is_done",
    "max_broken",
    "max_trials",
    "metadata",
    "name",
    "pool_size",
    "producer",
    "refers",
    "retrieve_result",
    "space",
    "stats",
    "to_pandas",
    "version",
    "working_dir",
]
read_write_only_methods = [
    "fix_lost_trials",
    "register_lie",
    "register_trial",
    "set_trial_status",
    "update_completed_trial",
]
execute_only_methods = [
    "reserve_trial",
]

ignore = ["non_branching_attrs", "mode", "node"]

dummy_trial = Trial(**_generate(base_trial, "status", value="reserved"))

stati = (["completed"] * 3) + (["reserved"] * 2)
trials = generate_trials(stati)
running_trial = Trial(**trials[-1])

kwargs = {
    "fetch_trials_by_status": {"status": "completed"},
    "get_trial": {"uid": 0},
    "retrieve_result": {"trial": dummy_trial},
    "register_lie": {"lying_trial": dummy_trial},
    "register_trial": {"trial": dummy_trial},
    "set_trial_status": {"trial": dummy_trial, "status": "interrupted"},
    "update_completed_trial": {"trial": running_trial},
}


def test_coverage_of_access_tests():
    """Make sure all attributes and methods are tested to avoid missing access-restricted methods"""
    # TODO, make sure we test all methods.
    all_items = filter(lambda a: not a.startswith("_"), dir(Experiment))
    assert set(all_items) == set(
        read_only_methods + read_write_only_methods + execute_only_methods + ignore
    )


def compare_supported(attr_name, restricted_exp, execution_exp):
    restricted_attr = getattr(restricted_exp, attr_name)
    execution_attr = getattr(execution_exp, attr_name)

    if inspect.ismethod(restricted_attr):
        restricted_attr = restricted_attr(**kwargs.get(attr_name, {}))
        execution_attr = execution_attr(**kwargs.get(attr_name, {}))

    if attr_name == "to_pandas":
        pandas.testing.assert_frame_equal(restricted_attr, execution_attr)
    else:
        assert restricted_attr == execution_attr


def compare_unsupported(attr_name, restricted_exp, execution_exp):
    restricted_attr = getattr(restricted_exp, attr_name)
    execution_attr = getattr(execution_exp, attr_name)

    # So far we only have restricted methods. We will need to modify this test if we ever add
    # restricted properties.
    assert inspect.ismethod(restricted_attr), attr_name

    execution_attr = execution_attr(**kwargs.get(attr_name, {}))
    with pytest.raises(UnsupportedOperation) as exc:
        restricted_attr = restricted_attr(**kwargs.get(attr_name, {}))
    assert exc.match(f"to execute `{attr_name}()")


def create_experiment(mode, space, algorithm):
    experiment = Experiment("supernaekei", mode=mode)
    experiment.space = space
    experiment.algorithms = algorithm
    experiment.max_broken = 5
    experiment.max_trials = 5
    return experiment


class TestReadOnly:
    """Test Experiment access rights in readonly mode"""

    def test_read_only_methods(self, space, algorithm):
        with OrionState(trials=trials) as cfg:
            read_only_exp = create_experiment("r", space, algorithm)
            execution_exp = create_experiment("x", space, algorithm)

            for method in read_only_methods:
                compare_supported(method, read_only_exp, execution_exp)

    def test_read_write_methods(self):
        with OrionState(trials=trials) as cfg:
            read_only_exp = create_experiment("r", space, algorithm)
            execution_exp = create_experiment("x", space, algorithm)
            for method in read_write_only_methods + execute_only_methods:
                compare_unsupported(method, read_only_exp, execution_exp)


class TestReadWriteOnly:
    """Test Experiment access rights in read/write only mode"""

    def test_read_only_methods(self, space, algorithm):
        with OrionState(trials=trials) as cfg:
            read_only_exp = create_experiment("w", space, algorithm)
            execution_exp = create_experiment("x", space, algorithm)

            for method in read_only_methods:
                compare_supported(method, read_only_exp, execution_exp)

    def test_execution_methods(self):
        with OrionState(trials=trials) as cfg:
            read_only_exp = create_experiment("w", space, algorithm)
            execution_exp = create_experiment("x", space, algorithm)
            for method in execute_only_methods:
                compare_unsupported(method, read_only_exp, execution_exp)



# -*- coding: utf-8 -*-
"""
Package-wide useful routines
============================

"""

import logging
import os
from abc import ABCMeta
from collections import defaultdict
from glob import glob
from importlib import import_module

import pkg_resources

log = logging.getLogger(__name__)


def nesteddict():
    """
    Define type of arbitrary nested defaultdicts
    Extend defaultdict to arbitrary nested levels.
    """
    return defaultdict(nesteddict)


def get_all_subclasses(parent):
    """Get set of subclasses recursively"""
    subclasses = set()
    for subclass in parent.__subclasses__():
        subclasses.add(subclass)
        subclasses |= get_all_subclasses(subclass)

    return subclasses


def get_all_types(parent_cls, cls_name):
    """Get all subclasses and lowercase subclass names"""
    types = list(get_all_subclasses(parent_cls))
    types = [class_ for class_ in types if class_.__name__ != cls_name]

    return {class_.__name__.lower(): class_ for class_ in types}


def _import_modules(cls):
    cls.modules = []
    base = import_module(cls.__base__.__module__)
    try:
        py_files = glob(os.path.abspath(os.path.join(base.__path__[0], "[A-Za-z]*.py")))
        py_mods = map(
            lambda x: "." + os.path.split(os.path.splitext(x)[0])[1], py_files
        )
        for py_mod in py_mods:
            cls.modules.append(import_module(py_mod, package=cls.__base__.__module__))
    except AttributeError:
        # This means that base class and implementations reside in a module
        # itself and not a subpackage.
        pass

    # Get types advertised through entry points!
    for entry_point in pkg_resources.iter_entry_points(cls.__name__):
        entry_point.load()
        log.debug(
            "Found a %s %s from distribution: %s=%s",
            entry_point.name,
            cls.__name__,
            entry_point.dist.project_name,
            entry_point.dist.version,
        )


def _set_typenames(cls):
    # Get types visible from base module or package, but internal
    cls.types.update(get_all_types(cls.__base__, cls.__name__))

    log.debug("Implementations found: %s", sorted(cls.types.keys()))


class Factory(ABCMeta):
    """Instantiate appropriate wrapper for the infrastructure based on input
    argument, ``of_type``.

    Attributes
    ----------
    types : dict of subclasses of ``cls.__base__``
       Updated to contain all possible implementations currently. Check out code.

    """

    def __init__(cls, names, bases, dictionary):
        """Search in directory for attribute names subclassing `bases[0]`"""
        super(Factory, cls).__init__(names, bases, dictionary)
        cls.types = {}
        try:
            _import_modules(cls)
        except ImportError:
            pass
        _set_typenames(cls)

    def __call__(cls, of_type, *args, **kwargs):
        """Create an object, instance of ``cls.__base__``, on first call.

        :param of_type: Name of class, subclass of ``cls.__base__``, wrapper
           of a database framework that will be instantiated on the first call.
        :param args: positional arguments to initialize ``cls.__base__``'s instance (if any)
        :param kwargs: keyword arguments to initialize ``cls.__base__``'s instance (if any)

        .. seealso::
           `Factory.types` keys for values of argument `of_type`.

        .. seealso::
           Attributes of ``cls.__base__`` and ``cls.__base__.__init__`` for
           values of `args` and `kwargs`.

        .. note:: New object is saved as `Factory`'s internal state.

        :return: The object which was created on the first call.
        """
        _import_modules(cls)
        _set_typenames(cls)

        for name, inherited_class in cls.types.items():
            if name == of_type.lower():
                return inherited_class(*args, **kwargs)

        error = "Could not find implementation of {0}, type = '{1}'".format(
            cls.__base__.__name__, of_type
        )
        error += "\nCurrently, there is an implementation for types:\n"
        error += str(sorted(cls.types.keys()))
        raise NotImplementedError(error)



"""
====================
Checkpointing trials
====================

.. hint::

    In short, you should use "{experiment.working_dir}/{trial.hash_params}" to set the path of
    the checkpointing file.

When using multi-fidelity algorithms such as Hyperband it is preferable to checkpoint the trials
to avoid starting training from scratch when resuming a trial. In this tutorial for instance,
hyperband will train VGG11 for 1 epoch, pick the best candidates and train them for 7 more epochs,
doing the same again for 30 epoch, and then 120 epochs. We want to resume training at last epoch
instead of starting from scratch.

Oríon provides a unique hash for trials that can be used to define the unique checkpoint file
path: ``trial.hash_params``. This can be used with the Python API as demonstrated in this example
or with :ref:`commandline_templates`.

With command line
-----------------

The example below is based on the Python API solely. It is also possible to do checkpointing
using the command line API. To this end, your script should accept an argument for the checkpoint
file path. Suppose this argument is ``--checkpoint``, you should call your script with the
following template.

::

    orion hunt -n <exp name>
        ./your_script.sh --checkpoint '{experiment.working_dir}/{trial.hash_params}'

Your script is reponsible to take this checkpoint path, resume from checkpoints or same
checkpoints.
We will demonstrate below how this can be done with PyTorch, but using Oríon's Python API.

Training code
-------------

We will first go through the training code piece by piece before tackling the hyperparameter
optimization.

First things first, the imports.

"""
import numpy

import torch
import torch.nn as nn
import torch.optim as optim
from torch.utils.data import SubsetRandomSampler

import torchvision
import torchvision.models as models
import torchvision.transforms as transforms

import os
import argparse

#%%
# We will use the data SubsetRandomSampler data loader from PyTorch to split
# the training set into a training and validation sets. We include test set
# here for completeness but won't use it in this example as we only need the training
# data and the validation data for the hyperparameter optimization.
# We use torchvision's transformers to apply the standard transformations on CIFAR10
# images, that is, random cropping, random horizontal flipping and normalization.


def build_data_loaders(batch_size, split_seed=1):
    normalize = [
        transforms.ToTensor(),
        transforms.Normalize((0.4914, 0.4822, 0.4465), (0.2023, 0.1994, 0.2010)),
    ]

    augment = [
        transforms.RandomCrop(32, padding=4),
        transforms.RandomHorizontalFlip(),
    ]

    train_set = torchvision.datasets.CIFAR10(
        root="./data",
        train=True,
        download=True,
        transform=transforms.Compose(augment + normalize),
    )
    valid_set = torchvision.datasets.CIFAR10(
        root="./data",
        train=True,
        download=True,
        transform=transforms.Compose(normalize),
    )
    test_set = torchvision.datasets.CIFAR10(
        root="./data",
        train=False,
        download=True,
        transform=transforms.Compose(normalize),
    )

    num_train = 45000
    # num_valid = 5000
    indices = numpy.arange(num_train)
    numpy.random.RandomState(split_seed).shuffle(indices)

    train_idx, valid_idx = indices[:num_train], indices[num_train:]
    train_sampler = SubsetRandomSampler(train_idx)
    valid_sampler = SubsetRandomSampler(valid_idx)

    train_loader = torch.utils.data.DataLoader(
        train_set, batch_size=batch_size, sampler=train_sampler, num_workers=5
    )
    valid_loader = torch.utils.data.DataLoader(
        train_set, batch_size=1000, sampler=train_sampler, num_workers=5
    )
    test_loader = torch.utils.data.DataLoader(
        test_set, batch_size=1000, shuffle=False, num_workers=5
    )

    return train_loader, valid_loader, test_loader


#%%
# Next, we write the function to save checkpoints. It is important to include
# not only the model in the checkpoint, but also the optimizer and the learning rate
# schedule when using one. In this example we will use the exponential learning rate schedule,
# so we checkpoint it. We save the current epoch as well so that we now where we resume from.


def save_checkpoint(checkpoint, model, optimizer, lr_scheduler, epoch):
    if checkpoint is None:
        return

    state = {
        "model": model.state_dict(),
        "optimizer": optimizer.state_dict(),
        "lr_scheduler": lr_scheduler.state_dict(),
        "epoch": epoch,
    }
    torch.save(state, f"{checkpoint}/checkpoint.pth")


#%%
# To resume from checkpoints, we simply restore the states of the model, optimizer and learning rate
# schedules based on the checkpoint file. If there is no checkpoint path or if the file does not
# exist, we return epoch 1 so that training starts from scratch. Otherwise we return the last
# trained epoch number found in checkpoint file.


def resume_from_checkpoint(checkpoint, model, optimizer, lr_scheduler):
    if checkpoint is None:
        return 1

    try:
        state_dict = torch.load(f"{checkpoint}/checkpoint.pth")
    except FileNotFoundError:
        return 1

    model.load_state_dict(state_dict["model"])
    optimizer.load_state_dict(state_dict["optimizer"])
    lr_scheduler.load_state_dict(state_dict["lr_scheduler"])
    return state_dict["epoch"] + 1  # Start from next epoch


#%%
# Then comes the training loop for one epoch.


def train(loader, device, model, optimizer, lr_scheduler, criterion):
    model.train()
    for batch_idx, (inputs, targets) in enumerate(loader):
        inputs, targets = inputs.to(device), targets.to(device)
        optimizer.zero_grad()
        outputs = model(inputs)
        loss = criterion(outputs, targets)
        loss.backward()
        optimizer.step()
        lr_scheduler.step()


#%%
# Finally the validation loop to compute the validation error rate.


def valid(loader, device, model):
    model.eval()
    correct = 0
    total = 0
    with torch.no_grad():
        for batch_idx, (inputs, targets) in enumerate(loader):
            inputs, targets = inputs.to(device), targets.to(device)
            outputs = model(inputs)

            _, predicted = outputs.max(1)
            total += targets.size(0)
            correct += predicted.eq(targets).sum().item()

    return 100.0 * (1 - correct / total)


#%%
# We combine all these functions into a main function for the whole training pipeline.
#
# .. note::
#
#     We set ``batch_size`` to 1024 by default, you may need to reduce it depending on your GPU.
#


def main(
    epochs=120,
    learning_rate=0.1,
    momentum=0.9,
    weight_decay=0,
    batch_size=1024,
    gamma=0.97,
    checkpoint=None,
):

    # We create the checkpointing folder if it does not exist.
    if checkpoint and not os.path.isdir(checkpoint):
        os.makedirs(checkpoint)

    device = "cuda" if torch.cuda.is_available() else "cpu"

    model = models.vgg11()
    model = model.to(device)

    # We define the training criterion, optimizer and learning rate scheduler
    criterion = nn.CrossEntropyLoss()
    optimizer = optim.SGD(
        model.parameters(),
        lr=learning_rate,
        momentum=momentum,
        weight_decay=weight_decay,
    )
    lr_scheduler = torch.optim.lr_scheduler.ExponentialLR(optimizer, gamma)

    # We restore the states of model, optimizer and learning rate scheduler if a checkpoint file is
    # available. This will return the last epoch number of the checkpoint or 1 if no checkpoint.
    start_epoch = resume_from_checkpoint(checkpoint, model, optimizer, lr_scheduler)

    # We build the data loaders. test_loader is here for completeness but won't be used.
    train_loader, valid_loader, test_loader = build_data_loaders(batch_size=batch_size)

    # If no training needed, because the trial was resumed from an epoch equal or greater to number
    # of epochs requested here (``epochs``).
    if start_epoch >= epochs + 1:
        return valid(valid_loader, device, model)

    # Training from last epoch until ``epochs + 1``, checkpointing at end of each epoch.
    for epoch in range(start_epoch, epochs + 1):
        print("epoch", epoch)
        train(train_loader, device, model, optimizer, lr_scheduler, criterion)
        valid_error_rate = valid(valid_loader, device, model)
        save_checkpoint(checkpoint, model, optimizer, lr_scheduler, epoch)

    return valid_error_rate


#%%
# You can test the training pipeline before working with the hyperparameter optimization.


main(epochs=4)


#%%
# HPO code
# --------
#
# We finally implement the hyperparameter optimization loop. We will use Hyperband
# with the number of epochs as the fidelity, using the prior ``fidelity(1, 120, base=4)``.
# Hyperband will thus train VGG11 for 1, 7, 30 and 120 epochs. To explore enough candidates
# at 120 epochs, we set Hyperband with 5 repetitions.
#
# In the optimization loop (``while not experiment.is_done``), we ask Oríon to suggest a new trial
# and then pass the hyperparameter values ``**trial.params`` to ``main()``, specifying the
# checkpoint file with ``f"{experiment.working_dir}/{trial.hash_params}"``.


from orion.client import build_experiment


def run_hpo():

    # Specify the database where the experiments are stored. We use a local PickleDB here.
    storage = {
        "type": "legacy",
        "database": {
            "type": "pickleddb",
            "host": "./db.pkl",
        },
    }

    # Load the data for the specified experiment
    experiment = build_experiment(
        "hyperband-cifar10",
        space={
            "epochs": "fidelity(1, 120, base=4)",
            "learning_rate": "loguniform(1e-5, 0.1)",
            "momentum": "uniform(0, 0.9)",
            "weight_decay": "loguniform(1e-10, 1e-2)",
            "gamma": "loguniform(0.97, 1)",
        },
        algorithms={
            "hyperband": {
                "seed": 1,
                "repetitions": 5,
            },
        },
        storage=storage,
    )

    trials = 1
    while not experiment.is_done:
        print("trial", trials)
        trial = experiment.suggest()
        if trial is None and experiment.is_done:
            break
        valid_error_rate = main(
            **trial.params, checkpoint=f"{experiment.working_dir}/{trial.hash_params}"
        )
        experiment.observe(trial, valid_error_rate, name="valid_error_rate")
        trials += 1


#%%
# Let's run the optimization now. You may want to reduce the maximum number of epochs in
# ``fidelity(1, 120, base=4)`` and set the number of ``repetitions`` to 1 to get results more
# quickly. With current configuration, this example takes 2 days to run on a Titan RTX.

experiment = run_hpo()

#%%
# Analysis
# --------
#
# That is all for the checkpointing example. We should nevertheless analyse the results
# before wrapping up this tutorial.
#
# We should first look at the :ref:`sphx_glr_auto_examples_plot_1_regret.py`
# to verify the optimization with Hyperband.

fig = experiment.plot.regret()
fig.show()

#%%
# .. This file is produced by docs/scripts/build_database_and_plots.py
#
# .. raw:: html
#     :file: ../_static/hyperband-cifar10_regret.html
#
#
# Moving the cursor over the points, we see that only a handful of trials
# lead to better results with 1 epoch. Otherwise, all other trials with validation error rate
# below 80% were trained for more than 1 epoch.
# The best found result is high, a validation accuracy 23.6%. With VGG11 we could expect to achieve
# lower than 10%. To see if the search space may be the issue, we first look at the
# :ref:`sphx_glr_auto_examples_plot_3_lpi.py`.

fig = experiment.plot.lpi()
fig.show()

#%%
# .. raw:: html
#     :file: ../_static/hyperband-cifar10_lpi.html
#
# The momentum and weight decay had very large priors, yet the different values
# had no important effect on the validation accuracy. We can ignore them.
# For the learning rate and for gamma
# it is worth looking at the :ref:`sphx_glr_auto_examples_plot_4_partial_dependencies.py`
# to see if the search space was perhaps too narrow or too large.

fig = experiment.plot.partial_dependencies(params=["gamma", "learning_rate"])
fig.show()

# sphinx_gallery_thumbnail_path = '_static/restart.png'

#%%
# .. This file is produced by docs/scripts/build_database_and_plots.py
#
# .. raw:: html
#     :file: ../_static/hyperband-cifar10_partial_dependencies_params.html
#
# The main culprit for the high validation error rate seems to be the wide prior for ``gamma``.
# Because of this Hyperband spent most of the computation time on bad ``gamma``s. This prior
# should be narrowed to ``uniform(0.995, 1)``.
# The prior for the learning rate could also be narrowed to ``loguniform(0.001, 0.1)`` to
# help the optimization.
#
# Note that Hyperband could also find better results without adjusting the search space, but
# it would required significantly more repetitions.



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Base definition of Assessment
==============================
"""

from abc import ABC, abstractmethod

from orion.core.utils import Factory


class BaseAssess(ABC):
    """Base class describing what an assessment can do.

    Parameters
    ----------
    task_num : int
       Number of experiment the assessment ask to run the corresponding task
    kwargs : dict
       Configurable parameters of the assessment, a particular assessment
       implementation can have its own parameters.
    """

    def __init__(self, task_num, **kwargs):
        self.task_number = task_num
        self._param_names = kwargs
        self._param_names["task_num"] = task_num

    @property
    def task_num(self):
        """Return the task number to run for this assessment"""
        return self.task_number

    @abstractmethod
    def analysis(self, task, experiments):
        """
        Generate a `plotly.graph_objects.Figure` to display the performance analysis
        based on the assessment purpose.

        task: str
            Name of the task
        experiments: list
            A list of (task_index, experiment), where task_index is the index of task to run for
            this assessment, and experiment is an instance of `orion.core.worker.experiment`.
        """
        pass

    @property
    def configuration(self):
        """Return the configuration of the assessment."""
        return {self.__class__.__qualname__: self._param_names}


# pylint: disable=too-few-public-methods,abstract-method
class BenchmarkAssessment(BaseAssess, metaclass=Factory):
    """Class used to inject dependency on an assessment implementation."""

    pass



# -*- coding: utf-8 -*-
# pylint:disable=too-many-lines
"""
Experiment wrapper client
=========================

Wraps the core Experiment object to provide further functionalities for the user
"""
import inspect
import logging
import traceback
from contextlib import contextmanager

import orion.core
import orion.core.utils.format_trials as format_trials
from orion.core.io.database import DuplicateKeyError
from orion.core.utils.exceptions import (
    BrokenExperiment,
    CompletedExperiment,
    InvalidResult,
    SampleTimeout,
    UnsupportedOperation,
    WaitingForTrials,
)
from orion.core.utils.flatten import flatten, unflatten
from orion.core.worker.trial import Trial, TrialCM
from orion.core.worker.trial_pacemaker import TrialPacemaker
from orion.executor.base import Executor
from orion.plotting.base import PlotAccessor
from orion.storage.base import FailedUpdate

log = logging.getLogger(__name__)


def reserve_trial(experiment, producer, _depth=1):
    """Reserve a new trial, or produce and reserve a trial if none are available."""
    log.debug("Trying to reserve a new trial to evaluate.")
    trial = experiment.reserve_trial()

    if trial is None and not producer.is_done:

        if _depth > 10:
            raise WaitingForTrials(
                "No trials are available at the moment "
                "wait for current trials to finish"
            )

        log.debug("#### Failed to pull a new trial from database.")

        log.debug("#### Fetch most recent completed trials and update algorithm.")
        producer.update()

        log.debug("#### Produce new trials.")
        producer.produce()

        return reserve_trial(experiment, producer, _depth=_depth + 1)

    return trial


# pylint: disable=too-many-public-methods
class ExperimentClient:
    """ExperimentClient providing all functionalities for the python API

    Note that the ExperimentClient is not meant to be instantiated by the user.
    Look at `orion.client.create_experiment` to build an ExperimentClient.

    Parameters
    ----------
    experiment: `orion.core.worker.experiment.Experiment`
        Experiment object serving for interaction with storage
    producer: `orion.core.worker.producer.Producer`
        Producer object used to produce new trials.

    """

    def __init__(self, experiment, producer, executor=None, heartbeat=None):
        self._experiment = experiment
        self._producer = producer
        self._pacemakers = {}
        if heartbeat is None:
            heartbeat = orion.core.config.worker.heartbeat
        self.heartbeat = heartbeat
        self.executor = executor or Executor(
            orion.core.config.worker.executor,
            n_workers=orion.core.config.worker.n_workers,
            **orion.core.config.worker.executor_configuration,
        )
        self.plot = PlotAccessor(self)

    ###
    # Attributes
    ###

    @property
    def name(self):
        """Return the name of the experiment in the database."""
        return self._experiment.name

    # pylint: disable=invalid-name
    @property
    def id(self):
        """Return the id of the experiment in the database."""
        return self._experiment.id

    @property
    def version(self):
        """Version of the experiment."""
        return self._experiment.version

    @property
    def max_trials(self):
        """Max-trials to execute before stopping the experiment."""
        return self._experiment.max_trials

    @property
    def max_broken(self):
        """Minimum number of broken trials before the experiment is considered broken."""
        return self._experiment.max_broken

    @property
    def metadata(self):
        """Metadata of the experiment."""
        return self._experiment.metadata

    @property
    def space(self):
        """Return problem's parameter `orion.algo.space.Space`."""
        return self._experiment.space

    @property
    def algorithms(self):
        """Algorithms of the experiment."""
        return self._experiment.algorithms

    @property
    def refers(self):
        """References to the experiment version control"""
        return self._experiment.refers

    @property
    def is_done(self):
        """Return True, if this experiment is considered to be finished.

        1. Count how many trials have been completed and compare with `max_trials`.
        2. Ask `algorithms` if they consider there is a chance for further improvement.
        """
        return self._experiment.is_done

    @property
    def is_broken(self):
        """Return True, if this experiment is considered to be broken.

        Count how many trials are broken and return True if that number has reached
        as given threshold.
        """
        return self._experiment.is_broken

    @property
    def configuration(self):
        """Return a copy of an `Experiment` configuration as a dictionary."""
        return self._experiment.configuration

    @property
    def stats(self):
        """Calculate a stats dictionary for this particular experiment.

        Returns
        -------
        stats : dict

        Stats
        -----
        trials_completed : int
           Number of completed trials
        best_trials_id : int
           Unique identifier of the :class:`orion.core.worker.trial.Trial` object in the database
           which achieved the best known objective result.
        best_evaluation : float
           Evaluation score of the best trial
        start_time : `datetime.datetime`
           When Experiment was first dispatched and started running.
        finish_time : `datetime.datetime`
           When Experiment reached terminating condition and stopped running.
        duration : `datetime.timedelta`
           Elapsed time.

        """
        return self._experiment.stats

    @property
    def node(self):
        """Node of the experiment in the version control tree."""
        return self._experiment.node

    @property
    def working_dir(self):
        """Working directory of the experiment."""
        return self._experiment.working_dir

    @property
    def producer(self):
        """Return the producer configuration of the experiment."""
        return self._experiment.producer

    @property
    def mode(self):
        """Return the access right of the experiment

        {'r': read, 'w': read/write, 'x': read/write/execute}
        """
        return self._experiment.mode

    ###
    # Rights
    ###

    def _check_if_writable(self):
        if self.mode == "r":
            calling_function = inspect.stack()[1].function
            raise UnsupportedOperation(
                f"ExperimentClient must have write rights to execute `{calling_function}()`"
            )

    def _check_if_executable(self):
        if self.mode != "x":
            calling_function = inspect.stack()[1].function
            raise UnsupportedOperation(
                f"ExperimentClient must have execution rights to execute `{calling_function}()`"
            )

    ###
    # Queries
    ###

    def to_pandas(self, with_evc_tree=False):
        """Builds a dataframe with the trials of the experiment

        Parameters
        ----------
        with_evc_tree: bool, optional
            Fetch all trials from the EVC tree.
            Default: False

        """
        return self._experiment.to_pandas(with_evc_tree=with_evc_tree)

    def fetch_trials(self, with_evc_tree=False):
        """Fetch all trials of the experiment

        Parameters
        ----------
        with_evc_tree: bool, optional
            Fetch all trials from the EVC tree.
            Default: False

        """
        return self._experiment.fetch_trials(with_evc_tree=with_evc_tree)

    def get_trial(self, trial=None, uid=None):
        """Fetch a single trial

        Parameters
        ----------
        trial: Trial, optional
           trial object to retrieve from the database

        uid: str, optional
            trial id used to retrieve the trial object

        Returns
        -------
        return none if the trial is not found,

        Raises
        ------
        UndefinedCall
            if both trial and uid are not set

        AssertionError
            if both trial and uid are provided and they do not match

        """
        return self._experiment.get_trial(trial, uid)

    def fetch_trials_by_status(self, status, with_evc_tree=False):
        """Fetch all trials with the given status

        Trials are sorted based on ``Trial.submit_time``

        :return: list of :class:`orion.core.worker.trial.Trial` objects
        """
        return self._experiment.fetch_trials_by_status(
            status, with_evc_tree=with_evc_tree
        )

    def fetch_noncompleted_trials(self, with_evc_tree=False):
        """Fetch non-completed trials of this `Experiment` instance.

        Trials are sorted based on ``Trial.submit_time``

        .. note::

            It will return all non-completed trials, including new, reserved, suspended,
            interrupted and broken ones.

        :return: list of non-completed :class:`orion.core.worker.trial.Trial` objects
        """
        return self._experiment.fetch_noncompleted_trials(with_evc_tree=with_evc_tree)

    ###
    # Actions
    ###

    # pylint: disable=unused-argument
    def insert(self, params, results=None, reserve=False):
        """Insert a new trial.

        Experiment must be in writable ('w') or executable ('x') mode.

        Parameters
        ----------
        params: dict
            Parameters of the new trial to add to the database. These parameters
            must comply with the space definition otherwise a ValueError will be raised.
        results: list, optional
            Results to be set for the new trial. Results must have the format
            {name: <str>: type: <'objective', 'constraint' or 'gradient'>, value=<float>} otherwise
            a ValueError will be raised.
            Note that passing results will mark the trial as completed and therefore cannot be
            reserved. The returned trial will have status 'completed'.
            If the results are invalid, the trial will still be inserted but reservation will be
            released.
        reserve: bool, optional
            If reserve=True, the inserted trial will be reserved. `reserve` cannot be True if
            `results` are given.
            Defaults to False.

        Returns
        -------
        `orion.core.worker.trial.Trial`
            The trial inserted in storage. If `reserve=True` and no results are given, the returned
            trial will be in a `reserved` status.

        Raises
        ------
        `ValueError`
            - If results are given and reserve=True
            - If params have invalid format
            - If results have invalid format
        `orion.core.io.database.DuplicateKeyError`
            - If a trial with identical params already exist for the current experiment.
        `orion.core.utils.exceptions.UnsupportedOperation`
            If the experiment was not loaded in writable mode.

        """
        self._check_if_writable()

        if results and reserve:
            raise ValueError(
                "Cannot observe a trial and reserve it. A trial with results has status "
                "`completed` and cannot be reserved."
            )
        trial = format_trials.dict_to_trial(params, self.space)
        try:
            self._experiment.register_trial(trial, status="reserved")
            self._maintain_reservation(trial)
        except DuplicateKeyError as e:
            message = (
                "A trial with params {} already exist for experiment {}-v{}".format(
                    params, self.name, self.version
                )
            )
            raise DuplicateKeyError(message) from e

        if results:
            try:
                self.observe(trial, results)
            except ValueError:
                self._release_reservation(trial)
                raise

            return trial

        if not reserve:
            self.release(trial)

        return trial

    def reserve(self, trial):
        """Reserve a trial.

        Experiment must be in executable ('x') mode.

        Set a trial status to reserve to ensure that concurrent process cannot work on it.
        Trials can only be reserved with status 'new', 'interrupted' or 'suspended'.

        Parameters
        ----------
        trial: `orion.core.worker.trial.Trial`
            Trial to reserve.

        Raises
        ------
        `RuntimeError`
            If trial is reserved by another process
        `ValueError`
            If the trial does not exist in storage.
        `orion.core.utils.exceptions.UnsupportedOperation`
            If the experiment was not loaded in executable mode.

        Notes
        -----
        When reserved, a :class:`TrialPacemaker <orion.core.worker.trial_pacemaker.TrialPacemaker>`
        is started to update an heartbeat in storage. The frequency of the heartbeat is configurable
        at creation of experiment or with ``orion.core.config.worker.heartbeat``.
        If the process terminates unexpectedly, the heartbeat will cease and remote processes
        may reset the status of the trial to 'interrupted' when the heartbeat has not been updated
        since twice the value of ``heartbeat``.

        """
        self._check_if_executable()

        if trial.status == "reserved" and trial.id in self._pacemakers:
            log.warning("Trial %s is already reserved.", trial.id)
            return
        elif trial.status == "reserved" and trial.id not in self._pacemakers:
            raise RuntimeError(
                "Trial {} is already reserved by another process.".format(trial.id)
            )
        try:
            self._experiment.set_trial_status(
                trial, "reserved", heartbeat=self.heartbeat
            )
        except FailedUpdate as e:
            if self.get_trial(trial) is None:
                raise ValueError(
                    "Trial {} does not exist in database.".format(trial.id)
                ) from e
            raise RuntimeError("Could not reserve trial {}.".format(trial.id)) from e

        self._maintain_reservation(trial)

    def release(self, trial, status="interrupted"):
        """Release a trial.

        Release the reservation and stop the heartbeat.

        Experiment must be in writable ('w') or executable ('x') mode.

        Parameters
        ----------
        trial: `orion.core.worker.trial.Trial`
            Trial to reserve.
        status: str, optional
            Set the trial to given status while releasing the reservation.
            Defaults to 'interrupted'.

        Raises
        ------
        `RuntimeError`
            If reservation of the trial has been lost prior to releasing it.
        `ValueError`
            If the trial does not exist in storage.
        `orion.core.utils.exceptions.UnsupportedOperation`
            If the experiment was not loaded in writable mode.

        """
        self._check_if_writable()

        current_status = trial.status
        raise_if_unreserved = True
        try:
            self._experiment.set_trial_status(trial, status, was="reserved")
        except FailedUpdate as e:
            if self.get_trial(trial) is None:
                raise ValueError(
                    "Trial {} does not exist in database.".format(trial.id)
                ) from e
            if current_status != "reserved":
                raise_if_unreserved = False
                raise RuntimeError(
                    "Trial {} was already released locally.".format(trial.id)
                ) from e

            raise RuntimeError(
                "Reservation for trial {} has been lost before release.".format(
                    trial.id
                )
            ) from e
        finally:
            self._release_reservation(trial, raise_if_unreserved=raise_if_unreserved)

    def suggest(self):
        """Suggest a trial to execute.

        Experiment must be in executable ('x') mode.

        If any trial is available (new or interrupted), it selects one and reserves it.
        Otherwise, the algorithm is used to generate a new trial that is registered in storage and
        reserved.

        Returns
        -------
        `orior.core.worker.trial.Trial`
            Reserved trial for execution.

        Raises
        ------
        :class:`orion.core.utils.exceptions.WaitingForTrials`
            if the experiment is not completed and algorithm needs to wait for some
            trials to complete before it can suggest new trials.

        :class:`orion.core.utils.exceptions.BrokenExperiment`
            if too many trials failed to run and the experiment cannot continue.
            This is determined by ``max_broken`` in the configuration of the experiment.

        :class:`orion.core.utils.exceptions.SampleTimeout`
            if the algorithm of the experiment could not sample new unique trials.

        :class:`orion.core.utils.exceptions.CompletedExperiment`
            if the experiment was completed and algorithm could not sample new trials.

        :class:`orion.core.utils.exceptions.UnsupportedOperation`
            If the experiment was not loaded in executable mode.

        """
        self._check_if_executable()

        if self.is_broken:
            raise BrokenExperiment("Trials failed too many times")

        if self.is_done:
            raise CompletedExperiment("Experiment is done, cannot sample more trials.")

        try:
            trial = reserve_trial(self._experiment, self._producer)

        except (WaitingForTrials, SampleTimeout) as e:
            if self.is_broken:
                raise BrokenExperiment("Trials failed too many times") from e

            raise e

        # This is to handle cases where experiment was completed during call to `reserve_trial`
        if trial is None:
            raise CompletedExperiment("Producer is done, cannot sample more trials.")

        self._maintain_reservation(trial)
        return TrialCM(self, trial)

    def observe(self, trial, results):
        """Observe trial results

        Experiment must be in executable ('x') mode.

        Parameters
        ----------
        trial: `orion.core.worker.trial.Trial`
            Reserved trial to observe.
        results: list
            Results to be set for the new trial. Results must have the format
            {name: <str>: type: <'objective', 'constraint' or 'gradient'>, value=<float>} otherwise
            a ValueError will be raised. If the results are invalid, the trial will not be released.

        Returns
        -------
        `orion.core.worker.trial.Trial`
            The trial inserted in storage. If `reserve=True` and no results are given, the returned
            trial will be in a `reserved` status.

        Raises
        ------
        `ValueError`
            - If results have invalid format
            - If the trial does not exist in storage.
        `RuntimeError`
            If reservation of the trial has been lost prior to releasing it.
        `orion.core.utils.exceptions.UnsupportedOperation`
            If the experiment was not loaded in executable mode.

        """
        self._check_if_executable()

        trial.results += [Trial.Result(**result) for result in results]
        raise_if_unreserved = True
        try:
            self._experiment.update_completed_trial(trial)
        except FailedUpdate as e:
            if self.get_trial(trial) is None:
                raise_if_unreserved = False
                raise ValueError(
                    "Trial {} does not exist in database.".format(trial.id)
                ) from e

            raise RuntimeError(
                "Reservation for trial {} has been lost.".format(trial.id)
            ) from e
        finally:
            self._release_reservation(trial, raise_if_unreserved=raise_if_unreserved)

    @contextmanager
    def tmp_executor(self, executor, **config):
        """Temporarily change the executor backend of the experiment client.

        Parameters
        ----------
        executor: str or :class:`orion.executor.base.Executor`
            The executor to use. If it is a ``str``, the provided ``config`` will be used
            to create the executor with ``Executor(executor, **config)``.
        **config:
            Configuration to use if ``executor`` is a ``str``.

        """
        if isinstance(executor, str):
            executor = Executor(executor, **config)
        old_executor = self.executor
        self.executor = executor
        with executor:
            yield self
        self.executor = old_executor

    def workon(
        self,
        fct,
        n_workers=None,
        max_trials=None,
        max_trials_per_worker=None,
        max_broken=None,
        trial_arg=None,
        on_error=None,
        **kwargs,
    ):
        """Optimize a given function

        Experiment must be in executable ('x') mode.

        Parameters
        ----------
        fct: callable
            Function to optimize. Must take arguments provided by trial.params. Additional constant
            parameter can be passed as ``**kwargs`` to `workon`. Function must return the final
            objective.
        n_workers: int, optional
            Number of workers to run in parallel. Defaults to value of global config.
        max_trials: int, optional
            Maximum number of trials to execute within ``workon``. If the experiment or algorithm
            reach status is_done before, the execution of ``workon`` terminates.
            Defaults to experiment's max trial. If ``max_trials`` is larger than
            ``experiment.max_trials``, the experiment will stop when reaching
            ``experiment.max_trials``.
        max_trials_per_worker: int, optional
            Maximum number of trials to execute within each worker. ``max_trials`` and
            ``experiment.max_trials`` have precedence. Defaults to global config.
        max_broken: int, optional
            Maximum number of broken trials to accept during `workon`. When this threshold is
            reached the function will raise :class:`orion.core.utils.exceptions.BrokenExperiment`.
            Defaults to value of global config.
        trial_arg: str, optional
            The function ``fct`` may support receiving the trial as an argument. This argument name
            can be specified with ``trial_arg``. If not defined (``None``), then only the
            hyperparameters will be passed to `fct`.
        on_error: callable, optional
            Callback that is executed if an error occur during the execution of ``fct``.
            The signature of the callback must be
            ``foo(ExperimentClient, Trial, Error, nb_of_worker_broken_trials)``.
            If the callblack returns False, the error will be ignored, otherwise it is counted
            for the threshold `max_broken`. In case of critical errors, you may also directly
            raise an error and force break out of ``workon``.
        **kwargs
            Constant argument to pass to `fct` in addition to trial.params. If values in kwargs are
            present in trial.params, the latter takes precedence.

        Raises
        ------
        :class:`orion.core.utils.exceptions.InvalidResult`
             If results returned by `fct` have invalid format.

        :class:`orion.core.utils.exceptions.WaitingForTrials`
            if the experiment is not completed and algorithm needs to wait for some
            trials to complete before it can suggest new trials.

        :class:`orion.core.utils.exceptions.BrokenExperiment`
            if too many trials failed to run and the experiment cannot continue.
            This is determined by ``max_broken`` in the configuration of the experiment.

        :class:`orion.core.utils.exceptions.SampleTimeout`
            if the algorithm of the experiment could not sample new unique points.

        :class:`orion.core.utils.exceptions.UnsupportedOperation`
            If the experiment was not loaded in executable mode.

        """
        self._check_if_executable()

        if n_workers is None:
            n_workers = orion.core.config.worker.n_workers

        if max_trials is None:
            max_trials = self.max_trials

        if max_trials_per_worker is None:
            max_trials_per_worker = orion.core.config.worker.max_trials

        if max_broken is None:
            max_broken = orion.core.config.worker.max_broken

        # Use worker's max_trials inside `exp.is_done` to reduce chance of
        # race condition for trials creation
        if self.max_trials > max_trials:
            self._experiment.max_trials = max_trials
            self._experiment.algorithms.algorithm.max_trials = max_trials

        trials = self.executor.wait(
            self.executor.submit(
                self._optimize,
                fct,
                max_trials_per_worker,
                max_broken,
                trial_arg,
                on_error,
                **kwargs,
            )
            for _ in range(n_workers)
        )

        return sum(trials)

    def _optimize(self, fct, max_trials, max_broken, trial_arg, on_error, **kwargs):
        worker_broken_trials = 0
        trials = 0
        kwargs = flatten(kwargs)
        max_trials = min(max_trials, self.max_trials)
        while not self.is_done and trials - worker_broken_trials < max_trials:
            try:
                with self.suggest() as trial:

                    kwargs.update(flatten(trial.params))

                    if trial_arg:
                        kwargs[trial_arg] = trial

                    try:
                        results = self.executor.wait(
                            [self.executor.submit(fct, **unflatten(kwargs))]
                        )[0]
                        self.observe(trial, results=results)
                    except (KeyboardInterrupt, InvalidResult):
                        raise
                    except BaseException as e:
                        if on_error is None or on_error(
                            self, trial, e, worker_broken_trials
                        ):
                            log.error(traceback.format_exc())
                            worker_broken_trials += 1
                        else:
                            log.error(str(e))
                            log.debug(traceback.format_exc())

                        if worker_broken_trials >= max_broken:
                            raise BrokenExperiment(
                                "Worker has reached broken trials threshold"
                            )
                        else:
                            self.release(trial, status="broken")
            except CompletedExperiment as e:
                log.warning(e)
                break

            trials += 1

        return trials

    def close(self):
        """Verify that no reserved trials are remaining.

        Experiment must be in executable ('x') mode.

        Raises
        ------
        `orion.core.utils.exceptions.UnsupportedOperation`
            If the experiment was not loaded in executable mode.

        """
        self._check_if_executable()

        if self._pacemakers:
            raise RuntimeError(
                "There is still reserved trials: {}\nRelease all trials before "
                "closing the client, using "
                "client.release(trial).".format(self._pacemakers.keys())
            )

    ###
    # Private
    ###

    def __repr__(self):
        """Represent the object as a string."""
        return "Experiment(name=%s, version=%s)" % (self.name, self.version)

    def _verify_reservation(self, trial):
        if trial.id not in self._pacemakers:
            raise RuntimeError(
                "Trial {} had no pacemakers. Was it reserved properly?".format(trial.id)
            )

        if self.get_trial(trial).status != "reserved":
            self._release_reservation(trial)
            raise RuntimeError(
                "Reservation for trial {} has been lost.".format(trial.id)
            )

    def _maintain_reservation(self, trial):
        self._pacemakers[trial.id] = TrialPacemaker(trial)
        self._pacemakers[trial.id].start()

    def _release_reservation(self, trial, raise_if_unreserved=True):
        if trial.id not in self._pacemakers:
            if raise_if_unreserved:
                raise RuntimeError(
                    "Trial {} had no pacemakers. Was it reserved properly?".format(
                        trial.id
                    )
                )
            else:
                return

        self._pacemakers.pop(trial.id).stop()
